#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import copy
import logging
from typing import Dict, List, Tuple, Union

import numpy as np
import pandas as pd
import scipy.spatial as sc
from scipy.spatial import ConvexHull
from skimage.measure import marching_cubes

from ..biomarkers.get_oriented_bound_box import rot_matrix, sig_proc_find_peaks


def get_mesh(mask: np.ndarray,
             res: Union[np.ndarray, List]) -> Tuple[np.ndarray,
                                                    np.ndarray,
                                                    np.ndarray]:
    """Compute Mesh.

    Note:
      Make sure the `mask` is padded with a layer of 0's in all
      dimensions to reduce potential isosurface computation errors.

    Args:
        mask (ndarray): Contains only 0's and 1's.
        res (ndarray or List): [a,b,c] vector specifying the resolution of the volume in mm.
            xyz resolution (world), or JIK resolution (intrinsic matlab).

    Returns:
        Tuple[np.ndarray, np.ndarray, np.ndarray]: 
            - Array of the [X,Y,Z] positions of the ROI.
            - Array of the spatial coordinates for `mask` unique mesh vertices.
            - Array of triangular faces via referencing vertex indices from vertices.
    """
    # Getting the grid of X,Y,Z positions, where the coordinate reference
    # system (0,0,0) is located at the upper left corner of the first voxel
    # (-0.5: half a voxel distance). For the whole volume defining the mask,
    # no matter if it is a 1 or a 0.
    mask = mask.copy()
    res = res.copy()

    x = res[0]*((np.arange(1, np.shape(mask)[0]+1))-0.5)
    y = res[1]*((np.arange(1, np.shape(mask)[1]+1))-0.5)
    z = res[2]*((np.arange(1, np.shape(mask)[2]+1))-0.5)
    X, Y, Z = np.meshgrid(x, y, z, indexing='ij')

    # Getting the isosurface of the mask
    vertices, faces, _, _ = marching_cubes(volume=mask, level=0.5, spacing=res)

    # Getting the X,Y,Z positions of the ROI (i.e. 1's) of the mask
    X = np.reshape(X, (np.size(X), 1), order='F')
    Y = np.reshape(Y, (np.size(Y), 1), order='F')
    Z = np.reshape(Z, (np.size(Z), 1), order='F')

    xyz = np.concatenate((X, Y, Z), axis=1)
    xyz = xyz[np.where(np.reshape(mask, np.size(mask), order='F') == 1)[0], :]

    return xyz, faces, vertices

def get_com(xgl_int: np.ndarray,
            xgl_morph: np.ndarray,
            xyz_int: np.ndarray,
            xyz_morph: np.ndarray) -> Union[float,
                                            np.ndarray]:
    """Calculates center of mass shift (in mm, since resolution is in mm).

    Note: 
        Row positions of "x_gl" and "xyz" must correspond for each point.
    
    Args:
        xgl_int (ndarray): Vector of intensity values in the volume to analyze 
            (only values in the intensity mask).
        xgl_morph (ndarray): Vector of intensity values in the volume to analyze 
            (only values in the morphological mask).
        xyz_int (ndarray): [n_points X 3] matrix of three column vectors, defining the [X,Y,Z]
            positions of the points in the ROI (1's) of the mask volume (In mm).
            (Mesh-based volume calculated from the ROI intensity mesh)
        xyz_morph (ndarray): [n_points X 3] matrix of three column vectors, defining the [X,Y,Z]
            positions of the points in the ROI (1's) of the mask volume (In mm).
            (Mesh-based volume calculated from the ROI morphological mesh)

    Returns:
        Union[float, np.ndarray]: The ROI volume centre of mass.

    """

    # Getting the geometric centre of mass
    n_v = np.size(xgl_morph)

    com_geom = np.sum(xyz_morph, 0)/n_v  # [1 X 3] vector

    # Getting the density centre of mass
    xyz_int[:, 0] = xgl_int*xyz_int[:, 0]
    xyz_int[:, 1] = xgl_int*xyz_int[:, 1]
    xyz_int[:, 2] = xgl_int*xyz_int[:, 2]
    com_gl = np.sum(xyz_int, 0)/np.sum(xgl_int, 0)  # [1 X 3] vector

    # Calculating the shift
    com = np.linalg.norm(com_geom - com_gl)

    return com

def get_area_dens_approx(a: float,
                         b: float,
                         c: float,
                         n: float) -> float:
    """Computes area density - minimum volume enclosing ellipsoid
    
    Args:
        a (float): Major semi-axis length.
        b (float): Minor semi-axis length.
        c (float): Least semi-axis length.
        n (int): Number of iterations.

    Returns:
        float: Area density - minimum volume enclosing ellipsoid.

    """
    alpha = np.sqrt(1 - b**2/a**2)
    beta = np.sqrt(1 - c**2/a**2)
    ab = alpha * beta
    point = (alpha**2+beta**2) / (2*ab)
    a_ell = 0

    for v in range(0, n+1):
        coef = [0]*v + [1]
        legen = np.polynomial.legendre.legval(x=point, c=coef)
        a_ell = a_ell + ab**v / (1-4*v**2) * legen

    a_ell = a_ell * 4 * np.pi * a * b

    return a_ell

def get_axis_lengths(xyz: np.ndarray) -> Tuple[float, float, float]:
    """Computes AxisLengths.
    
    Args:
        xyz (ndarray): Array of three column vectors, defining the [X,Y,Z]
            positions of the points in the ROI (1's) of the mask volume. In mm.

    Returns:
        Tuple[float, float, float]: Array of three column vectors 
        [Major axis lengths, Minor axis lengths, Least axis lengths].

    """
    xyz = xyz.copy()

    # Getting the geometric centre of mass
    com_geom = np.sum(xyz, 0)/np.shape(xyz)[0]  # [1 X 3] vector

    # Subtracting the centre of mass
    xyz[:, 0] = xyz[:, 0] - com_geom[0]
    xyz[:, 1] = xyz[:, 1] - com_geom[1]
    xyz[:, 2] = xyz[:, 2] - com_geom[2]

    # Getting the covariance matrix
    cov_mat = np.cov(xyz, rowvar=False)

    # Getting the eigenvalues
    eig_val, _ = np.linalg.eig(cov_mat)
    eig_val = np.sort(eig_val)

    major = eig_val[2]
    minor = eig_val[1]
    least = eig_val[0]

    return major, minor, least

def min_oriented_bound_box(pos_mat: np.ndarray) -> np.ndarray:
    """Computes the minimum bounding box of an arbitrary solid: an iterative approach.
    This feature refers to "Volume density (oriented minimum bounding box)" (ID = ZH1A)  
    in the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`_.

    Args:
        pos_mat (ndarray): matrix position

    Returns:
        ndarray: return bounding box dimensions
    """

    ##########################
    # Internal functions
    ##########################

    def calc_rot_aabb_surface(theta: float,
                              hull_mat: np.ndarray) -> np.ndarray:
        """Function to calculate surface of the axis-aligned bounding box of a rotated 2D contour

        Args:
            theta (float): angle in radian
            hull_mat (nddarray): convex hull matrix

        Returns:
            ndarray: the surface of the axis-aligned bounding box of a rotated 2D contour
        """

        # Create rotation matrix and rotate over theta
        rot_mat = rot_matrix(theta=theta, dim=2)
        rot_hull = np.dot(rot_mat, hull_mat)

        # Calculate bounding box surface of the rotated contour
        rot_aabb_dims = np.max(rot_hull, axis=1) - np.min(rot_hull, axis=1)
        rot_aabb_area = np.product(rot_aabb_dims)

        return rot_aabb_area

    def approx_min_theta(hull_mat: np.ndarray,
                         theta_sel: float,
                         res: float,
                         max_rep: int=5) -> np.ndarray:
        """Iterative approximator for finding angle theta that minimises surface area

        Args:
            hull_mat (ndarray): convex hull matrix
            theta_sel (float): angle in radian
            res (float): value in radian
            max_rep (int, optional): maximum repetition. Defaults to 5.

        Returns:
            ndarray: the angle theta that minimises surfae area
        """

        for i in np.arange(0, max_rep):

            # Select new thetas in vicinity of
            theta = np.array([theta_sel-res, theta_sel-0.5*res,
                              theta_sel, theta_sel+0.5*res, theta_sel+res])

            # Calculate projection areas for current angles theta
            rot_area = np.array(
                list(map(lambda x: calc_rot_aabb_surface(theta=x, hull_mat=hull_mat), theta)))

            # Find global minimum and corresponding angle theta_sel
            theta_sel = theta[np.argmin(rot_area)]

            # Shrink resolution and iterate
            res = res / 2.0

        return theta_sel

    def rotate_minimal_projection(input_pos: float,
                                  rot_axis: int,
                                  n_minima: int=3,
                                  res_init: float=5.0):
        """Function to that rotates input_pos to find the rotation that
        minimises the projection of input_pos on the
        plane normal to the rot_axis

        Args:
            input_pos (float): input position value
            rot_axis (int): rotation axis value
            n_minima (int, optional): _description_. Defaults to 3.
            res_init (float, optional): _description_. Defaults to 5.0.

        Returns:
            _type_: _description_
        """


        # Find axis aligned bounding box of the point set
        aabb_max = np.max(input_pos, axis=0)
        aabb_min = np.min(input_pos, axis=0)

        # Center the point set at the AABB center
        output_pos = input_pos - 0.5 * (aabb_min + aabb_max)

        # Project model to plane
        proj_pos = copy.deepcopy(output_pos)
        proj_pos = np.delete(proj_pos, [rot_axis], axis=1)

        # Calculate 2D convex hull of the model projection in plane
        if np.shape(proj_pos)[0] >= 10:
            hull_2d = ConvexHull(points=proj_pos)
            hull_mat = proj_pos[hull_2d.vertices, :]
            del hull_2d, proj_pos
        else:
            hull_mat = proj_pos
            del proj_pos

        # Transpose hull_mat so that the array is (ndim, npoints) instead of (npoints, ndim)
        hull_mat = np.transpose(hull_mat)

        # Calculate bounding box surface of a series of rotated contours
        # Note we can program a min-search algorithm here as well

        # Calculate initial surfaces
        theta_init = np.arange(start=0.0, stop=90.0 +
                               res_init, step=res_init) * np.pi / 180.0
        rot_area = np.array(
            list(map(lambda x: calc_rot_aabb_surface(theta=x, hull_mat=hull_mat), theta_init)))

        # Find local minima
        df_min = sig_proc_find_peaks(x=rot_area, ddir="neg")

        # Check if any minimum was generated
        if len(df_min) > 0:
            # Investigate up to n_minima number of local minima, starting with the global minimum
            df_min = df_min.sort_values(by="val", ascending=True)

            # Determine max number of minima evaluated
            max_iter = np.min([n_minima, len(df_min)])

            # Initialise place holder array
            theta_min = np.zeros(max_iter)

            # Iterate over local minima
            for k in np.arange(0, max_iter):

                # Find initial angle corresponding to i-th minimum
                sel_ind = df_min.ind.values[k]
                theta_curr = theta_init[sel_ind]

                # Zoom in to improve the approximation of theta
                theta_min[k] = approx_min_theta(
                    hull_mat=hull_mat, theta_sel=theta_curr, res=res_init*np.pi/180.0)

            # Calculate surface areas corresponding to theta_min and theta that
            # minimises the surface
            rot_area = np.array(
                list(map(lambda x: calc_rot_aabb_surface(theta=x, hull_mat=hull_mat), theta_min)))
            theta_sel = theta_min[np.argmin(rot_area)]

        else:
            theta_sel = theta_init[0]

        # Rotate original point along the angle that minimises the projected AABB area
        output_pos = np.transpose(output_pos)
        rot_mat = rot_matrix(theta=theta_sel, dim=3, rot_axis=rot_axis)
        output_pos = np.dot(rot_mat, output_pos)

        # Rotate output_pos back to (npoints, ndim)
        output_pos = np.transpose(output_pos)

        return output_pos

    ##########################
    # Main function
    ##########################

    rot_df = pd.DataFrame({"rot_axis_0":  np.array([0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2]),
                           "rot_axis_1":  np.array([1, 2, 1, 2, 0, 2, 0, 2, 0, 1, 0, 1]),
                           "rot_axis_2":  np.array([2, 1, 0, 0, 2, 0, 1, 1, 1, 0, 2, 2]),
                           "aabb_axis_0": np.zeros(12),
                           "aabb_axis_1": np.zeros(12),
                           "aabb_axis_2": np.zeros(12),
                           "vol":         np.zeros(12)})

    # Rotate over different sequences
    for i in np.arange(0, len(rot_df)):
        # Create a local copy
        work_pos = copy.deepcopy(pos_mat)

        # Rotate over sequence of rotation axes
        work_pos = rotate_minimal_projection(
            input_pos=work_pos, rot_axis=rot_df.rot_axis_0[i])
        work_pos = rotate_minimal_projection(
            input_pos=work_pos, rot_axis=rot_df.rot_axis_1[i])
        work_pos = rotate_minimal_projection(
            input_pos=work_pos, rot_axis=rot_df.rot_axis_2[i])

        # Determine resultant minimum bounding box
        aabb_dims = np.max(work_pos, axis=0) - np.min(work_pos, axis=0)
        rot_df.loc[i, "aabb_axis_0"] = aabb_dims[0]
        rot_df.loc[i, "aabb_axis_1"] = aabb_dims[1]
        rot_df.loc[i, "aabb_axis_2"] = aabb_dims[2]
        rot_df.loc[i, "vol"] = np.product(aabb_dims)

        del work_pos, aabb_dims

    # Find minimal volume of all rotations and return bounding box dimensions
    idxmin = rot_df.vol.idxmin()
    sel_row = rot_df.loc[idxmin]
    ombb_dims = np.array(
        [sel_row.aabb_axis_0, sel_row.aabb_axis_1, sel_row.aabb_axis_2])

    return ombb_dims

def get_moran_i(vol: np.ndarray,
                res: List[float]) -> float:
    """Computes Moran's Index.
    This feature refers to "Moran’s I index" (ID = N365)  
    in the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`_.

    Args:
        vol (ndarray): 3D volume, NON-QUANTIZED, continous imaging intensity distribution.
        res (List[float]): [a,b,c] vector specfying the resolution of the volume in mm.
            XYZ resolution (world) or JIK resolution (intrinsic matlab).

    Returns:
        float: Value of Moran's Index.

    """
    vol = vol.copy()
    res = res.copy()

    # Find the location(s) of all non NaNs voxels
    I, J, K = np.nonzero(~np.isnan(vol))
    n_vox = np.size(I)

    # Get the mean
    u = np.mean(vol[~np.isnan(vol[:])])
    vol_mean = vol.copy() - u  # (x_gl,i - u)
    vol_m_mean_s = np.power((vol.copy() - u), 2)  # (x_gl,i - u).^2
    # Sum of (x_gl,i - u).^2 over all i
    sum_s = np.sum(vol_m_mean_s[~np.isnan(vol_m_mean_s[:])])

    # Get a meshgrid first
    x = res[0]*((np.arange(1, np.shape(vol)[0]+1))-0.5)
    y = res[1]*((np.arange(1, np.shape(vol)[1]+1))-0.5)
    z = res[2]*((np.arange(1, np.shape(vol)[2]+1))-0.5)
    X, Y, Z = np.meshgrid(x, y, z, indexing='ij')

    temp = 0
    sum_w = 0
    for i in range(1, n_vox+1):
        # Distance mesh
        temp_x = X - X[I[i-1], J[i-1], K[i-1]]
        temp_y = Y - Y[I[i-1], J[i-1], K[i-1]]
        temp_z = Z - Z[I[i-1], J[i-1], K[i-1]]

        # meshgrid of weigths
        temp_dist_mesh = 1 / np.sqrt(temp_x**2 + temp_y**2 + temp_z**2)

        # Removing NaNs
        temp_dist_mesh[np.isnan(vol)] = np.NaN
        temp_dist_mesh[I[i-1], J[i-1], K[i-1]] = np.NaN
        # Running sum of weights
        w_sum = np.sum(temp_dist_mesh[~np.isnan(temp_dist_mesh[:])])
        sum_w = sum_w + w_sum

        # Inside sum calculation
        # Removing NaNs
        temp_vol = vol_mean.copy()
        temp_vol[I[i-1], J[i-1], K[i-1]] = np.NaN
        temp_vol = temp_dist_mesh * temp_vol  # (wij .* (x_gl,j - u))
        # Summing (wij .* (x_gl,j - u)) over all j
        sum_val = np.sum(temp_vol[~np.isnan(temp_vol[:])])
        # Running sum of (x_gl,i - u)*(wij .* (x_gl,j - u)) over all i
        temp = temp + vol_mean[I[i-1], J[i-1], K[i-1]] * sum_val

    moran_i = temp*n_vox/sum_s/sum_w

    return moran_i

def get_mesh_volume(faces: np.ndarray,
                    vertices:np.ndarray) -> float:
    """Computes MeshVolume feature.
    This feature refers to "Volume (mesh)" (ID = RNU0)  
    in the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`_.

    Args:
        faces (np.ndarray): matrix of three column vectors, defining the [X,Y,Z]
                          positions of the ``faces`` of the isosurface or convex hull of the mask
                          (output from "isosurface.m" or "convhull.m" functions of MATLAB).
                          --> These are more precisely indexes to ``vertices``
        vertices (np.ndarray): matrix of three column vectors, defining the
                             [X,Y,Z] positions of the ``vertices`` of the isosurface of the mask (output
                             from "isosurface.m" function of MATLAB).
                             --> In mm.

    Returns:
        float: Mesh volume
    """
    faces = faces.copy()
    vertices = vertices.copy()

    # Getting vectors for the three vertices
    # (with respect to origin) of each face
    a = vertices[faces[:, 0], :]
    b = vertices[faces[:, 1], :]
    c = vertices[faces[:, 2], :]

    # Calculating volume
    v_cross = np.cross(b, c)
    v_dot = np.sum(a.conj()*v_cross, axis=1)
    volume = np.abs(np.sum(v_dot))/6

    return volume

def get_mesh_area(faces: np.ndarray,
                  vertices: np.ndarray) -> float:
    """Computes the surface area (mesh) feature from the ROI mesh by 
    summing over the triangular face surface areas. 
    This feature refers to "Surface area (mesh)" (ID = C0JK)  
    in the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`_.

    Args:
        faces (np.ndarray): matrix of three column vectors, defining the [X,Y,Z]
                          positions of the ``faces`` of the isosurface or convex hull of the mask
                          (output from "isosurface.m" or "convhull.m" functions of MATLAB).
                          --> These are more precisely indexes to ``vertices``
        vertices (np.ndarray): matrix of three column vectors,
                             defining the [X,Y,Z]
                             positions of the ``vertices`` of the isosurface of the mask (output
                             from "isosurface.m" function of MATLAB).
                             --> In mm.

    Returns:
        float: Mesh area.
    """

    faces = faces.copy()
    vertices = vertices.copy()

    # Getting two vectors of edges for each face
    a = vertices[faces[:, 1], :] - vertices[faces[:, 0], :]
    b = vertices[faces[:, 2], :] - vertices[faces[:, 0], :]

    # Calculating the surface area of each face and summing it up all at once.
    c = np.cross(a, b)
    area = 1/2 * np.sum(np.sqrt(np.sum(np.power(c, 2), 1)))

    return area

def get_geary_c(vol: np.ndarray,
                res: np.ndarray) -> float:
    """Computes Geary'C measure (Assesses intensity differences between voxels).
    This feature refers to "Geary's C measure" (ID = NPT7) 
    in the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`_.

    Args:
        vol (ndarray): 3D volume, NON-QUANTIZED, continous imaging intensity distribution.
        res (ndarray): [a,b,c] vector specfying the resolution of the volume in mm.

    Returns:
        float: computes value of Geary'C measure.
    """
    vol = vol.copy()
    res = res.copy()

    # Find the location(s) of all non NaNs voxels
    I, J, K = np.nonzero(~np.isnan(vol))
    n_vox = np.size(I)

    # Get the mean
    u = np.mean(vol[~np.isnan(vol[:])])
    vol_m_mean_s = np.power((vol.copy() - u), 2)  # (x_gl,i - u).^2

    # Sum of (x_gl,i - u).^2 over all i
    sum_s = np.sum(vol_m_mean_s[~np.isnan(vol_m_mean_s[:])])

    # Get a meshgrid first
    x = res[0]*((np.arange(1, np.shape(vol)[0]+1))-0.5)
    y = res[1]*((np.arange(1, np.shape(vol)[1]+1))-0.5)
    z = res[2]*((np.arange(1, np.shape(vol)[2]+1))-0.5)
    X, Y, Z = np.meshgrid(x, y, z, indexing='ij')

    temp = 0
    sum_w = 0

    for i in range(1, n_vox+1):
        # Distance mesh
        temp_x = X - X[I[i-1], J[i-1], K[i-1]]
        temp_y = Y - Y[I[i-1], J[i-1], K[i-1]]
        temp_z = Z - Z[I[i-1], J[i-1], K[i-1]]

        # meshgrid of weigths
        temp_dist_mesh = 1/np.sqrt(temp_x**2 + temp_y**2 + temp_z**2)

        # Removing NaNs
        temp_dist_mesh[np.isnan(vol)] = np.NaN
        temp_dist_mesh[I[i-1], J[i-1], K[i-1]] = np.NaN

        # Running sum of weights
        w_sum = np.sum(temp_dist_mesh[~np.isnan(temp_dist_mesh[:])])
        sum_w = sum_w + w_sum

        # Inside sum calculation
        val = vol[I[i-1], J[i-1], K[i-1]].copy()  # x_gl,i
        # wij.*(x_gl,i - x_gl,j).^2
        temp_vol = temp_dist_mesh*(vol - val)**2

        # Removing i voxel to be sure;
        temp_vol[I[i-1], J[i-1], K[i-1]] = np.NaN

        # Sum of wij.*(x_gl,i - x_gl,j).^2 over all j
        sum_val = np.sum(temp_vol[~np.isnan(temp_vol[:])])

        # Running sum of (sum of wij.*(x_gl,i - x_gl,j).^2 over all j) over all i
        temp = temp + sum_val

    geary_c = temp * (n_vox-1) / sum_s / (2*sum_w)

    return geary_c

def min_vol_ellipse(P: np.ndarray,
                    tolerance: np.ndarray) -> Tuple[np.ndarray,
                                                    np.ndarray]:
    """Computes min_vol_ellipse.
    
    Finds the minimum volume enclsing ellipsoid (MVEE) of a set of data
    points stored in matrix P. The following optimization problem is solved:

        minimize $$log(det(A))$$ subject to $$(P_i - c)' * A * (P_i - c) <= 1$$

    in variables A and c, where `P_i` is the `i-th` column of the matrix `P`.
    The solver is based on Khachiyan Algorithm, and the final solution
    is different from the optimal value by the pre-spesified amount of
    `tolerance`.
    
    Note:
        Adapted from MATLAB code of Nima Moshtagh (nima@seas.upenn.edu)
        University of Pennsylvania.

    Args:
        P (ndarray): (d x N) dimnesional matrix containing N points in R^d.
        tolerance (ndarray): error in the solution with respect to the optimal value.
    
    Returns:
        2-element tuple containing
    
        - A: (d x d) matrix of the ellipse equation in the 'center form': \
        $$(x-c)' * A * (x-c) = 1$$ \
        where d is shape of `P` along 0-axis. 
        
        - c: d-dimensional vector as the center of the ellipse. 

    Examples:

        >>>P = rand(5,100)

        >>>[A, c] = :func:`min_vol_ellipse(P, .01)`

        To reduce the computation time, work with the boundary points only:

        >>>K = :func:`convhulln(P)`

        >>>K = :func:`unique(K(:))`

        >>>Q = :func:`P(:,K)`

        >>>[A, c] = :func:`min_vol_ellipse(Q, .01)`
    """

    # Solving the Dual problem
    # data points
    d, N = np.shape(P)
    Q = np.ones((d+1, N))
    Q[:-1, :] = P[:, :]

    # initializations
    err = 1
    u = np.ones(N)/N  # 1st iteration
    new_u = np.zeros(N)

    # Khachiyan Algorithm

    while (err > tolerance):
        diag_u = np.diag(u)
        trans_q = np.transpose(Q)
        X = Q @ diag_u @ trans_q

        # M the diagonal vector of an NxN matrix
        inv_x = np.linalg.inv(X)
        M = np.diag(trans_q @ inv_x @ Q)
        maximum = np.max(M)
        j = np.argmax(M)

        step_size = (maximum - d - 1)/((d+1)*(maximum-1))
        new_u = (1 - step_size)*u.copy()
        new_u[j] = new_u[j] + step_size
        err = np.linalg.norm(new_u - u)
        u = new_u.copy()

    # Computing the Ellipse parameters
    # Finds the ellipse equation in the 'center form':
    # (x-c)' * A * (x-c) = 1
    # It computes a dxd matrix 'A' and a d dimensional vector 'c' as the center
    # of the ellipse.
    U = np.diag(u)

    # the A matrix for the ellipse
    c = P @ u
    c = np.reshape(c, (np.size(c), 1), order='F')  # center of the ellipse

    pup_t = P @ U @ np.transpose(P)
    cct = c @ np.transpose(c)
    a_inv = np.linalg.inv(pup_t - cct)
    A = (1/d) * a_inv

    return A, c

def padding(vol: np.ndarray, 
            mask_int: np.ndarray, 
            mask_morph: np.ndarray) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Padding the volume and masks.

    Args:
        vol (ndarray): 3D volume, NON-QUANTIZED, continous imaging intensity distribution.
        mask_int (ndarray): Intensity mask.
        mask_morph (ndarray): Morphological mask.

    Returns:
        tuple of 3 ndarray: Volume and masks after padding.
    """

    # PADDING THE VOLUME WITH A LAYER OF NaNs
    # (reduce mesh computation errors of associated mask)
    vol = vol.copy()
    vol = np.pad(vol, pad_width=1, mode="constant", constant_values=np.NaN)
    # PADDING THE MASKS WITH A LAYER OF 0's
    # (reduce mesh computation errors of associated mask)
    mask_int = mask_int.copy()
    mask_int = np.pad(mask_int, pad_width=1, mode="constant", constant_values=0.0)
    mask_morph = mask_morph.copy()
    mask_morph = np.pad(mask_morph, pad_width=1, mode="constant", constant_values=0.0)

    return vol, mask_int, mask_morph

def get_variables(vol: np.ndarray, 
                  mask_int: np.ndarray, 
                  mask_morph: np.ndarray,
                  res: np.ndarray) -> Tuple[np.ndarray, 
                                            np.ndarray, 
                                            np.ndarray, 
                                            np.ndarray, 
                                            np.ndarray, 
                                            np.ndarray, 
                                            np.ndarray]:
    """Compute variables usefull to calculate morphological features.

    Args:
        vol (ndarray): 3D volume, NON-QUANTIZED, continous imaging intensity distribution.
        mask_int (ndarray): Intensity mask.
        mask_morph (ndarray): Morphological mask.
        res (ndarray): [a,b,c] vector specfying the resolution of the volume in mm.
            XYZ resolution (world), or JIK resolution (intrinsic matlab).

    Returns:
        tuple of 7 ndarray: Variables usefull to calculate morphological features.
    """
    # GETTING IMPORTANT VARIABLES
    xgl_int = np.reshape(vol, np.size(vol), order='F')[np.where(
        np.reshape(mask_int, np.size(mask_int), order='F') == 1)[0]].copy()
    xgl_morph = np.reshape(vol, np.size(vol), order='F')[np.where(
        np.reshape(mask_morph, np.size(mask_morph), order='F') == 1)[0]].copy()
    # XYZ refers to [Xc,Yc,Zc] in ref. [1].
    xyz_int, _, _ = get_mesh(mask_int, res)
    # XYZ refers to [Xc,Yc,Zc] in ref. [1].
    xyz_morph, faces, vertices = get_mesh(mask_morph, res)
    # [X,Y,Z] points of the convex hull.
    # conv_hull Matlab is conv_hull.simplices
    conv_hull = sc.ConvexHull(vertices)

    return xgl_int, xgl_morph, xyz_int, xyz_morph, faces, vertices, conv_hull

def extract_all(vol: np.ndarray, 
                mask_int: np.ndarray, 
                mask_morph: np.ndarray, 
                res: np.ndarray, 
                intensity_type: str,
                compute_moran_i: bool=False, 
                compute_geary_c: bool=False) -> Dict:
    """Compute Morphological Features.
    This features refer to Morphological family in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.
    
    Note:
        Moran's Index and Geary's C measure takes so much computation time. Please
        use `compute_moran_i` `compute_geary_c` carefully.

    Args:
        vol (ndarray): 3D volume, NON-QUANTIZED, continous imaging intensity distribution.
        mask_int (ndarray): Intensity mask.
        mask_morph (ndarray): Morphological mask.
        res (ndarray): [a,b,c] vector specfying the resolution of the volume in mm.
            XYZ resolution (world), or JIK resolution (intrinsic matlab).
        intensity_type (str): Type of intensity to compute. Can be "arbitrary", "definite" or "filtered".
            Will compute all features for "definite" intensity type and all but intergrated intensity for 
            "arbitrary" intensity type.
        compute_moran_i (bool, optional): True to compute Moran's Index.
        compute_geary_c (bool, optional): True to compute Geary's C measure.

    Raises:
        ValueError: If `intensity_type` is not "arbitrary", "definite" or "filtered".
    """
    assert intensity_type in ["arbitrary", "definite", "filtered"], \
        "intensity_type must be 'arbitrary', 'definite' or 'filtered'"
    
    # Initialization of final structure (Dictionary) containing all features.
    morph = {'Fmorph_vol': [],
             'Fmorph_approx_vol': [],
             'Fmorph_area': [],
             'Fmorph_av': [],
             'Fmorph_comp_1': [],
             'Fmorph_comp_2': [],
             'Fmorph_sph_dispr': [],
             'Fmorph_sphericity': [],
             'Fmorph_asphericity': [],
             'Fmorph_com': [],
             'Fmorph_diam': [],
             'Fmorph_pca_major': [],
             'Fmorph_pca_minor': [],
             'Fmorph_pca_least': [],
             'Fmorph_pca_elongation': [],
             'Fmorph_pca_flatness': [],  # until here
             'Fmorph_v_dens_aabb': [],
             'Fmorph_a_dens_aabb': [],
             'Fmorph_v_dens_ombb': [],
             'Fmorph_a_dens_ombb': [],
             'Fmorph_v_dens_aee': [],
             'Fmorph_a_dens_aee': [],
             'Fmorph_v_dens_mvee': [],
             'Fmorph_a_dens_mvee': [],
             'Fmorph_v_dens_conv_hull': [],
             'Fmorph_a_dens_conv_hull': [],
             'Fmorph_integ_int': [],
             'Fmorph_moran_i': [],
             'Fmorph_geary_c': []
            }
    #Initialization
    vol, mask_int, mask_morph = padding(vol, mask_int, mask_morph)
    xgl_int, xgl_morph, xyz_int, xyz_morph, faces, vertices, conv_hull = get_variables(vol, mask_int, mask_morph, res)

    # STARTING COMPUTATION
    if intensity_type != "filtered":
        # Volume in mm^3
        volume = get_mesh_volume(faces, vertices)
        morph['Fmorph_vol'] = volume  # Volume

        # Approximate Volume
        morph['Fmorph_approx_vol'] = np.sum(mask_morph[:]) * np.prod(res)

        # Surface area in mm^2
        area = get_mesh_area(faces, vertices)
        morph['Fmorph_area'] = area

        # Surface to volume ratio
        morph['Fmorph_av'] = area / volume

        # Compactness 1
        morph['Fmorph_comp_1'] = volume / ((np.pi**(1/2))*(area**(3/2)))

        # Compactness 2
        morph['Fmorph_comp_2'] = 36*np.pi*(volume**2) / (area**3)

        # Spherical disproportion
        morph['Fmorph_sph_dispr'] = area / (36*np.pi*volume**2)**(1/3)

        # Sphericity
        morph['Fmorph_sphericity'] = ((36*np.pi*volume**2)**(1/3)) / area

        # Asphericity
        morph['Fmorph_asphericity'] = ((area**3) / (36*np.pi*volume**2))**(1/3) - 1

        # Centre of mass shift
        morph['Fmorph_com'] = get_com(xgl_int, xgl_morph, xyz_int, xyz_morph)

        # Maximum 3D diameter
        morph['Fmorph_diam'] = np.max(sc.distance.pdist(conv_hull.points[conv_hull.vertices]))

        # Major axis length
        [major, minor, least] = get_axis_lengths(xyz_morph)
        morph['Fmorph_pca_major'] = 4 * np.sqrt(major)

        # Minor axis length
        morph['Fmorph_pca_minor'] = 4 * np.sqrt(minor)

        # Least axis length
        morph['Fmorph_pca_least'] = 4 * np.sqrt(least)

        # Elongation
        morph['Fmorph_pca_elongation'] = np.sqrt(minor / major)

        # Flatness
        morph['Fmorph_pca_flatness'] = np.sqrt(least / major)

        # Volume density - axis-aligned bounding box
        xc_aabb = np.max(vertices[:, 0]) - np.min(vertices[:, 0])
        yc_aabb = np.max(vertices[:, 1]) - np.min(vertices[:, 1])
        zc_aabb = np.max(vertices[:, 2]) - np.min(vertices[:, 2])
        v_aabb = xc_aabb * yc_aabb * zc_aabb
        morph['Fmorph_v_dens_aabb'] = volume / v_aabb

        # Area density - axis-aligned bounding box
        a_aabb = 2*xc_aabb*yc_aabb + 2*xc_aabb*zc_aabb + 2*yc_aabb*zc_aabb
        morph['Fmorph_a_dens_aabb'] = area / a_aabb

        # Volume density - oriented minimum bounding box
        # Implementation of Chan and Tan's algorithm (C.K. Chan, S.T. Tan.
        # Determination of the minimum bounding box of an
        # arbitrary solid: an iterative approach.
        # Comp Struc 79 (2001) 1433-1449
        bound_box_dims = min_oriented_bound_box(vertices)
        vol_bb = np.prod(bound_box_dims)
        morph['Fmorph_v_dens_ombb'] = volume / vol_bb

        # Area density - oriented minimum bounding box
        a_ombb = 2 * (bound_box_dims[0]*bound_box_dims[1] +
                        bound_box_dims[0]*bound_box_dims[2] +
                        bound_box_dims[1]*bound_box_dims[2])
        morph['Fmorph_a_dens_ombb'] = area / a_ombb

        # Volume density - approximate enclosing ellipsoid
        a = 2*np.sqrt(major)
        b = 2*np.sqrt(minor)
        c = 2*np.sqrt(least)
        v_aee = (4*np.pi*a*b*c) / 3
        morph['Fmorph_v_dens_aee'] = volume / v_aee

        # Area density - approximate enclosing ellipsoid
        a_aee = get_area_dens_approx(a, b, c, 20)
        morph['Fmorph_a_dens_aee'] = area / a_aee

        # Volume density - minimum volume enclosing ellipsoid
        # (Rotate the volume first??)
        # Copyright (c) 2009, Nima Moshtagh
        # http://www.mathworks.com/matlabcentral/fileexchange/
        # 9542-minimum-volume-enclosing-ellipsoid
        # Subsequent singular value decomposition of matrix A and and
        # taking the inverse of the square root of the diagonal of the
        # sigma matrix will produce respective semi-axis lengths.
        # Subsequent singular value decomposition of matrix A and
        # taking the inverse of the square root of the diagonal of the
        # sigma matrix will produce respective semi-axis lengths.
        p = np.stack((conv_hull.points[conv_hull.simplices[:, 0], 0],
                        conv_hull.points[conv_hull.simplices[:, 1], 1],
                        conv_hull.points[conv_hull.simplices[:, 2], 2]), axis=1)
        A, _ = min_vol_ellipse(np.transpose(p), 0.01)
        # New semi-axis lengths
        _, Q, _ = np.linalg.svd(A)
        a = 1/np.sqrt(Q[2])
        b = 1/np.sqrt(Q[1])
        c = 1/np.sqrt(Q[0])
        v_mvee = (4*np.pi*a*b*c)/3
        morph['Fmorph_v_dens_mvee'] = volume / v_mvee

        # Area density - minimum volume enclosing ellipsoid
        # Using a new set of (a,b,c), see Volume density - minimum
        # volume enclosing ellipsoid
        a_mvee = get_area_dens_approx(a, b, c, 20)
        morph['Fmorph_a_dens_mvee'] = area / a_mvee

        # Volume density - convex hull
        v_convex = conv_hull.volume
        morph['Fmorph_v_dens_conv_hull'] = volume / v_convex

        # Area density - convex hull
        a_convex = conv_hull.area
        morph['Fmorph_a_dens_conv_hull'] = area / a_convex

    # Integrated intensity
    if intensity_type == "definite":
        volume = get_mesh_volume(faces, vertices)
        morph['Fmorph_integ_int'] = np.mean(xgl_int) * volume

    # Moran's I index
    if compute_moran_i:
        vol_mor = vol.copy()
        vol_mor[mask_int == 0] = np.NaN
        morph['Fmorph_moran_i'] = get_moran_i(vol_mor, res)

    # Geary's C measure
    if compute_geary_c:
        morph['Fmorph_geary_c'] = get_geary_c(vol_mor, res)

    return morph

def vol(vol: np.ndarray, 
        mask_int: np.ndarray, 
        mask_morph: np.ndarray, 
        res: np.ndarray) -> float:
    """Computes morphological volume feature.
    This feature refers to "Fmorph_vol" (ID = RNUO) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        vol (ndarray): 3D volume, NON-QUANTIZED, continous imaging intensity distribution.
        mask_int (ndarray): Intensity mask.
        mask_morph (ndarray): Morphological mask.
        res (ndarray): [a,b,c] vector specfying the resolution of the volume in mm.
            XYZ resolution (world), or JIK resolution (intrinsic matlab).

    Returns:
        float: Value of the morphological volume feature.
    """
    vol, mask_int, mask_morph = padding(vol, mask_int, mask_morph)
    _, _, _, _, faces, vertices, _ = get_variables(vol, mask_int, mask_morph, res)
    volume = get_mesh_volume(faces, vertices)

    return volume  # Morphological volume feature

def approx_vol(vol: np.ndarray, 
               mask_int: np.ndarray, 
               mask_morph: np.ndarray, 
               res: np.ndarray) -> float:
    """Computes morphological approximate volume feature.
    This feature refers to "Fmorph_approx_vol" (ID = YEKZ) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        vol (ndarray): 3D volume, NON-QUANTIZED, continous imaging intensity distribution.
        mask_int (ndarray): Intensity mask.
        mask_morph (ndarray): Morphological mask.
        res (ndarray): [a,b,c] vector specfying the resolution of the volume in mm.
                       XYZ resolution (world), or JIK resolution (intrinsic matlab).

    Returns:
        float: Value of the morphological approximate volume feature.
    """
    vol, mask_int, mask_morph = padding(vol, mask_int, mask_morph)
    volume_appro = np.sum(mask_morph[:]) * np.prod(res)

    return volume_appro  # Morphological approximate volume feature

def area(vol: np.ndarray, 
         mask_int: np.ndarray, 
         mask_morph: np.ndarray, 
         res: np.ndarray) -> float:
    """Computes Surface area feature.
    This feature refers to "Fmorph_area" (ID = COJJK) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        vol (ndarray): 3D volume, NON-QUANTIZED, continous imaging intensity distribution.
        mask_int (ndarray): Intensity mask.
        mask_morph (ndarray): Morphological mask.
        res (ndarray): [a,b,c] vector specfying the resolution of the volume in mm.
                       XYZ resolution (world), or JIK resolution (intrinsic matlab).

    Returns:
        float: Value of the surface area feature.
    """
    vol, mask_int, mask_morph = padding(vol, mask_int, mask_morph)
    _, _, _, _, faces, vertices, _ = get_variables(vol, mask_int, mask_morph, res)
    area = get_mesh_area(faces, vertices)

    return area  # Surface area

def av(vol: np.ndarray, 
       mask_int: np.ndarray, 
       mask_morph: np.ndarray, 
       res: np.ndarray) -> float:
    """Computes Surface to volume ratio feature.
    This feature refers to "Fmorph_av" (ID = 2PR5) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        vol (ndarray): 3D volume, NON-QUANTIZED, continous imaging intensity distribution.
        mask_int (ndarray): Intensity mask.
        mask_morph (ndarray): Morphological mask.
        res (ndarray): [a,b,c] vector specfying the resolution of the volume in mm.
                       XYZ resolution (world), or JIK resolution (intrinsic matlab).

    Returns:
        float: Value of the Surface to volume ratio feature.
    """
    _, _, _, _, faces, vertices, _ = get_variables(vol, mask_int, mask_morph, res)
    volume = get_mesh_volume(faces, vertices)
    area = get_mesh_area(faces, vertices)
    ratio = area / volume

    return ratio  # Surface to volume ratio

def comp_1(vol: np.ndarray, 
           mask_int: np.ndarray, 
           mask_morph: np.ndarray, 
           res: np.ndarray) -> float:
    """Computes Compactness 1 feature.
    This feature refers to "Fmorph_comp_1" (ID = SKGS) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        vol (ndarray): 3D volume, NON-QUANTIZED, continous imaging intensity distribution.
        mask_int (ndarray): Intensity mask.
        mask_morph (ndarray): Morphological mask.
        res (ndarray): [a,b,c] vector specfying the resolution of the volume in mm.
                       XYZ resolution (world), or JIK resolution (intrinsic matlab).

    Returns:
        float: Value of the Compactness 1 feature.
    """
    _, _, _, _, faces, vertices, _ = get_variables(vol, mask_int, mask_morph, res)
    volume = get_mesh_volume(faces, vertices)
    area = get_mesh_area(faces, vertices)
    comp_1 = volume / ((np.pi**(1/2))*(area**(3/2)))

    return comp_1  # Compactness 1

def comp_2(vol: np.ndarray, 
           mask_int: np.ndarray, 
           mask_morph: np.ndarray, 
           res: np.ndarray) -> float:
    """Computes Compactness 2 feature.
    This feature refers to "Fmorph_comp_2" (ID = BQWJ) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        vol (ndarray): 3D volume, NON-QUANTIZED, continous imaging intensity distribution.
        mask_int (ndarray): Intensity mask.
        mask_morph (ndarray): Morphological mask.
        res (ndarray): [a,b,c] vector specfying the resolution of the volume in mm.
                       XYZ resolution (world), or JIK resolution (intrinsic matlab).

    Returns:
        float: Value of the Compactness 2 feature.
    """
    _, _, _, _, faces, vertices, _ = get_variables(vol, mask_int, mask_morph, res)
    volume = get_mesh_volume(faces, vertices)
    area = get_mesh_area(faces, vertices)
    comp_2 = 36*np.pi*(volume**2) / (area**3)

    return comp_2  # Compactness 2

def sph_dispr(vol: np.ndarray, 
              mask_int: np.ndarray, 
              mask_morph: np.ndarray, 
              res: np.ndarray) -> float:
    """Computes Spherical disproportion feature.
    This feature refers to "Fmorph_sph_dispr" (ID = KRCK) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        vol (ndarray): 3D volume, NON-QUANTIZED, continous imaging intensity distribution.
        mask_int (ndarray): Intensity mask.
        mask_morph (ndarray): Morphological mask.
        res (ndarray): [a,b,c] vector specfying the resolution of the volume in mm.
                       XYZ resolution (world), or JIK resolution (intrinsic matlab).

    Returns:
        float: Value of the Spherical disproportion feature.
    """
    _, _, _, _, faces, vertices, _ = get_variables(vol, mask_int, mask_morph, res)
    volume = get_mesh_volume(faces, vertices)
    area = get_mesh_area(faces, vertices)
    sph_dispr = area / (36*np.pi*volume**2)**(1/3)

    return sph_dispr  # Spherical disproportion

def sphericity(vol: np.ndarray, 
               mask_int: np.ndarray, 
               mask_morph: np.ndarray, 
               res: np.ndarray) -> float:
    """Computes Sphericity feature.
    This feature refers to "Fmorph_sphericity" (ID = QCFX) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        vol (ndarray): 3D volume, NON-QUANTIZED, continous imaging intensity distribution.
        mask_int (ndarray): Intensity mask.
        mask_morph (ndarray): Morphological mask.
        res (ndarray): [a,b,c] vector specfying the resolution of the volume in mm.
                       XYZ resolution (world), or JIK resolution (intrinsic matlab).

    Returns:
        float: Value of the Sphericity feature.
    """
    _, _, _, _, faces, vertices, _ = get_variables(vol, mask_int, mask_morph, res)
    volume = get_mesh_volume(faces, vertices)
    area = get_mesh_area(faces, vertices)
    sphericity = ((36*np.pi*volume**2)**(1/3)) / area

    return sphericity  # Sphericity

def asphericity(vol: np.ndarray, 
                mask_int: np.ndarray, 
                mask_morph: np.ndarray, 
                res: np.ndarray) -> float:
    """Computes Asphericity feature.
    This feature refers to "Fmorph_asphericity" (ID =  25C) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        vol (ndarray): 3D volume, NON-QUANTIZED, continous imaging intensity distribution.
        mask_int (ndarray): Intensity mask.
        mask_morph (ndarray): Morphological mask.
        res (ndarray): [a,b,c] vector specfying the resolution of the volume in mm.
                       XYZ resolution (world), or JIK resolution (intrinsic matlab).

    Returns:
        float: Value of the Asphericity feature.
    """
    _, _, _, _, faces, vertices, _ = get_variables(vol, mask_int, mask_morph, res)
    volume = get_mesh_volume(faces, vertices)
    area = get_mesh_area(faces, vertices)
    asphericity = ((area**3) / (36*np.pi*volume**2))**(1/3) - 1

    return asphericity  # Asphericity

def com(vol: np.ndarray, 
        mask_int: np.ndarray, 
        mask_morph: np.ndarray, 
        res: np.ndarray) -> float:
    """Computes Centre of mass shift feature.
    This feature refers to "Fmorph_com" (ID =  KLM) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        vol (ndarray): 3D volume, NON-QUANTIZED, continous imaging intensity distribution.
        mask_int (ndarray): Intensity mask.
        mask_morph (ndarray): Morphological mask.
        res (ndarray): [a,b,c] vector specfying the resolution of the volume in mm.
                       XYZ resolution (world), or JIK resolution (intrinsic matlab).

    Returns:
        float: Value of the Centre of mass shift feature.
    """
    vol, mask_int, mask_morph = padding(vol, mask_int, mask_morph)
    xgl_int, xgl_morph, xyz_int, xyz_morph, _, _, _ = get_variables(vol, mask_int, mask_morph, res)
    com = get_com(xgl_int, xgl_morph, xyz_int, xyz_morph)

    return com  # Centre of mass shift

def diam(vol: np.ndarray, 
         mask_int: np.ndarray, 
         mask_morph: np.ndarray, 
         res: np.ndarray) -> float:
    """Computes Maximum 3D diameter feature.
    This feature refers to "Fmorph_diam" (ID = L0JK) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        vol (ndarray): 3D volume, NON-QUANTIZED, continous imaging intensity distribution.
        mask_int (ndarray): Intensity mask.
        mask_morph (ndarray): Morphological mask.
        res (ndarray): [a,b,c] vector specfying the resolution of the volume in mm.
                       XYZ resolution (world), or JIK resolution (intrinsic matlab).

    Returns:
        float: Value of the Maximum 3D diameter feature.
    """
    vol, mask_int, mask_morph = padding(vol, mask_int, mask_morph)
    _, _, _, _, _, _, conv_hull = get_variables(vol, mask_int, mask_morph, res)
    diam = np.max(sc.distance.pdist(conv_hull.points[conv_hull.vertices]))

    return diam  # Maximum 3D diameter

def pca_major(vol: np.ndarray, 
              mask_int: np.ndarray, 
              mask_morph: np.ndarray, 
              res: np.ndarray) -> float:
    """Computes Major axis length feature.
    This feature refers to "Fmorph_pca_major" (ID = TDIC) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        vol (ndarray): 3D volume, NON-QUANTIZED, continous imaging intensity distribution.
        mask_int (ndarray): Intensity mask.
        mask_morph (ndarray): Morphological mask.
        res (ndarray): [a,b,c] vector specfying the resolution of the volume in mm.
                    XYZ resolution (world), or JIK resolution (intrinsic matlab).

    Returns:
        float: Value of the Major axis length feature.
    """
    vol, mask_int, mask_morph = padding(vol, mask_int, mask_morph)
    _, _, _, xyz_morph, _, _, _ = get_variables(vol, mask_int, mask_morph, res)
    [major, _, _] = get_axis_lengths(xyz_morph)
    pca_major = 4 * np.sqrt(major)

    return pca_major  # Major axis length

def pca_minor(vol: np.ndarray, 
              mask_int: np.ndarray, 
              mask_morph: np.ndarray, 
              res: np.ndarray) -> float:
    """Computes Minor axis length feature.
    This feature refers to "Fmorph_pca_minor" (ID = P9VJ) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        vol (ndarray): 3D volume, NON-QUANTIZED, continous imaging intensity distribution.
        mask_int (ndarray): Intensity mask.
        mask_morph (ndarray): Morphological mask.
        res (ndarray): [a,b,c] vector specfying the resolution of the volume in mm.
                       XYZ resolution (world), or JIK resolution (intrinsic matlab).

    Returns:
        float: Value of the Minor axis length feature.
    """
    vol, mask_int, mask_morph = padding(vol, mask_int, mask_morph)
    _, _, _, xyz_morph, _, _, _ = get_variables(vol, mask_int, mask_morph, res)
    [_, minor, _] = get_axis_lengths(xyz_morph)
    pca_minor = 4 * np.sqrt(minor)

    return pca_minor  # Minor axis length

def pca_least(vol: np.ndarray, 
              mask_int: np.ndarray, 
              mask_morph: np.ndarray, 
              res: np.ndarray) -> float:
    """Computes Least axis length feature.
    This feature refers to "Fmorph_pca_least" (ID = 7J51) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        vol (ndarray): 3D volume, NON-QUANTIZED, continous imaging intensity distribution.
        mask_int (ndarray): Intensity mask.
        mask_morph (ndarray): Morphological mask.
        res (ndarray): [a,b,c] vector specfying the resolution of the volume in mm.
                       XYZ resolution (world), or JIK resolution (intrinsic matlab).

    Returns:
        float: Value of the Least axis length feature.
    """
    vol, mask_int, mask_morph = padding(vol, mask_int, mask_morph)
    _, _, _, xyz_morph, _, _, _ = get_variables(vol, mask_int, mask_morph, res)
    [_, _, least] = get_axis_lengths(xyz_morph)
    pca_least = 4 * np.sqrt(least)

    return pca_least  # Least axis length

def pca_elongation(vol: np.ndarray, 
                   mask_int: np.ndarray, 
                   mask_morph: np.ndarray, 
                   res: np.ndarray) -> float:
    """Computes Elongation feature.
    This feature refers to "Fmorph_pca_elongation" (ID = Q3CK) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        vol (ndarray): 3D volume, NON-QUANTIZED, continous imaging intensity distribution.
        mask_int (ndarray): Intensity mask.
        mask_morph (ndarray): Morphological mask.
        res (ndarray): [a,b,c] vector specfying the resolution of the volume in mm.
                       XYZ resolution (world), or JIK resolution (intrinsic matlab).

    Returns:
        float: Value of the Elongation feature.
    """
    vol, mask_int, mask_morph = padding(vol, mask_int, mask_morph)
    _, _, _, xyz_morph, _, _, _ = get_variables(vol, mask_int, mask_morph, res)
    [major, minor, _] = get_axis_lengths(xyz_morph)
    pca_elongation = np.sqrt(minor / major)

    return pca_elongation  # Elongation

def pca_flatness(vol: np.ndarray, 
                 mask_int: np.ndarray, 
                 mask_morph: np.ndarray, 
                 res: np.ndarray) -> float:
    """Computes Flatness feature.
    This feature refers to "Fmorph_pca_flatness" (ID = N17B) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        vol (ndarray): 3D volume, NON-QUANTIZED, continous imaging intensity distribution.
        mask_int (ndarray): Intensity mask.
        mask_morph (ndarray): Morphological mask.
        res (ndarray): [a,b,c] vector specfying the resolution of the volume in mm.
                       XYZ resolution (world), or JIK resolution (intrinsic matlab).

    Returns:
        float: Value of the Flatness feature.
    """
    vol, mask_int, mask_morph = padding(vol, mask_int, mask_morph)
    _, _, _, xyz_morph, _, _, _ = get_variables(vol, mask_int, mask_morph, res)
    [major, _, least] = get_axis_lengths(xyz_morph)
    pca_flatness = np.sqrt(least / major)

    return pca_flatness  # Flatness

def v_dens_aabb(vol: np.ndarray, 
                mask_int: np.ndarray, 
                mask_morph: np.ndarray, 
                res: np.ndarray) -> float:
    """Computes Volume density - axis-aligned bounding box feature.
    This feature refers to "Fmorph_v_dens_aabb" (ID = PBX1) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        vol (ndarray): 3D volume, NON-QUANTIZED, continous imaging intensity distribution.
        mask_int (ndarray): Intensity mask.
        mask_morph (ndarray): Morphological mask.
        res (ndarray): [a,b,c] vector specfying the resolution of the volume in mm.
                       XYZ resolution (world), or JIK resolution (intrinsic matlab).

    Returns:
        float: Value of the Volume density - axis-aligned bounding box feature.
    """
    vol, mask_int, mask_morph = padding(vol, mask_int, mask_morph)
    _, _, _, _, faces, vertices, _ = get_variables(vol, mask_int, mask_morph, res)
    volume = get_mesh_volume(faces, vertices)
    xc_aabb = np.max(vertices[:, 0]) - np.min(vertices[:, 0])
    yc_aabb = np.max(vertices[:, 1]) - np.min(vertices[:, 1])
    zc_aabb = np.max(vertices[:, 2]) - np.min(vertices[:, 2])
    v_aabb = xc_aabb * yc_aabb * zc_aabb
    v_dens_aabb = volume / v_aabb

    return v_dens_aabb  # Volume density - axis-aligned bounding box

def a_dens_aabb(vol: np.ndarray, 
                mask_int: np.ndarray, 
                mask_morph: np.ndarray, 
                res: np.ndarray) -> float:
    """Computes Area density - axis-aligned bounding box feature.
    This feature refers to "Fmorph_a_dens_aabb" (ID = R59B) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        vol (ndarray): 3D volume, NON-QUANTIZED, continous imaging intensity distribution.
        mask_int (ndarray): Intensity mask.
        mask_morph (ndarray): Morphological mask.
        res (ndarray): [a,b,c] vector specfying the resolution of the volume in mm.
                       XYZ resolution (world), or JIK resolution (intrinsic matlab).

    Returns:
        float: Value of the Area density - axis-aligned bounding box feature.
    """
    vol, mask_int, mask_morph = padding(vol, mask_int, mask_morph)
    _, _, _, _, faces, vertices, _ = get_variables(vol, mask_int, mask_morph, res)
    area = get_mesh_area(faces, vertices)
    xc_aabb = np.max(vertices[:, 0]) - np.min(vertices[:, 0])
    yc_aabb = np.max(vertices[:, 1]) - np.min(vertices[:, 1])
    zc_aabb = np.max(vertices[:, 2]) - np.min(vertices[:, 2])
    a_aabb = 2*xc_aabb*yc_aabb + 2*xc_aabb*zc_aabb + 2*yc_aabb*zc_aabb
    a_dens_aabb = area / a_aabb

    return a_dens_aabb  # Area density - axis-aligned bounding box

def v_dens_ombb(vol: np.ndarray, 
                mask_int: np.ndarray, 
                mask_morph: np.ndarray, 
                res: np.ndarray) -> float:
    """Computes Volume density - oriented minimum bounding box feature.
    Implementation of Chan and Tan's algorithm (C.K. Chan, S.T. Tan.
    Determination of the minimum bounding box of an
    arbitrary solid: an iterative approach.
    Comp Struc 79 (2001) 1433-1449.
    This feature refers to "Fmorph_v_dens_ombb" (ID = ZH1A) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        vol (ndarray): 3D volume, NON-QUANTIZED, continous imaging intensity distribution.
        mask_int (ndarray): Intensity mask.
        mask_morph (ndarray): Morphological mask.
        res (ndarray): [a,b,c] vector specfying the resolution of the volume in mm.
                       XYZ resolution (world), or JIK resolution (intrinsic matlab).

    Returns:
        float: Value of the Volume density - oriented minimum bounding box feature.
    """
    vol, mask_int, mask_morph = padding(vol, mask_int, mask_morph)
    _, _, _, _, faces, vertices, _ = get_variables(vol, mask_int, mask_morph, res)
    volume = get_mesh_volume(faces, vertices)
    bound_box_dims = min_oriented_bound_box(vertices)
    vol_bb = np.prod(bound_box_dims)
    v_dens_ombb = volume / vol_bb

    return v_dens_ombb  # Volume density - oriented minimum bounding box

def a_dens_ombb(vol: np.ndarray, 
                mask_int: np.ndarray, 
                mask_morph: np.ndarray, 
                res: np.ndarray) -> float:
    """Computes Area density - oriented minimum bounding box feature.
    Implementation of Chan and Tan's algorithm (C.K. Chan, S.T. Tan.
    Determination of the minimum bounding box of an
    arbitrary solid: an iterative approach.
    This feature refers to "Fmorph_a_dens_ombb" (ID = IQYR) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        vol (ndarray): 3D volume, NON-QUANTIZED, continous imaging intensity distribution.
        mask_int (ndarray): Intensity mask.
        mask_morph (ndarray): Morphological mask.
        res (ndarray): [a,b,c] vector specfying the resolution of the volume in mm.
                       XYZ resolution (world), or JIK resolution (intrinsic matlab).

    Returns:
        float: Value of the Area density - oriented minimum bounding box feature.
    """
    vol, mask_int, mask_morph = padding(vol, mask_int, mask_morph)
    _, _, _, _, faces, vertices, _ = get_variables(vol, mask_int, mask_morph, res)
    area = get_mesh_area(faces, vertices)

    bound_box_dims = min_oriented_bound_box(vertices)
    a_ombb = 2 * (bound_box_dims[0] * bound_box_dims[1] 
                + bound_box_dims[0] * bound_box_dims[2]
                + bound_box_dims[1] * bound_box_dims[2])
    a_dens_ombb = area / a_ombb

    return a_dens_ombb  # Area density - oriented minimum bounding box

def v_dens_aee(vol: np.ndarray, 
               mask_int: np.ndarray, 
               mask_morph: np.ndarray, 
               res: np.ndarray) -> float:
    """Computes Volume density - approximate enclosing ellipsoid feature.
    This feature refers to "Fmorph_v_dens_aee" (ID = 6BDE) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args: 
        vol (ndarray): 3D volume, NON-QUANTIZED, continous imaging intensity distribution.
        mask_int (ndarray): Intensity mask.
        mask_morph (ndarray): Morphological mask.
        res (ndarray): [a,b,c] vector specfying the resolution of the volume in mm.
                       XYZ resolution (world), or JIK resolution (intrinsic matlab).

    Returns:
        float: Value of the Volume density - approximate enclosing ellipsoid feature.
    """
    vol, mask_int, mask_morph = padding(vol, mask_int, mask_morph)
    _, _, _, xyz_morph, faces, vertices, _ = get_variables(vol, mask_int, mask_morph, res)
    volume = get_mesh_volume(faces, vertices)
    [major, minor, least] = get_axis_lengths(xyz_morph)
    a = 2*np.sqrt(major)
    b = 2*np.sqrt(minor)
    c = 2*np.sqrt(least)
    v_aee = (4*np.pi*a*b*c) / 3
    v_dens_aee = volume / v_aee

    return v_dens_aee  # Volume density - approximate enclosing ellipsoid

def a_dens_aee(vol: np.ndarray, 
               mask_int: np.ndarray, 
               mask_morph: np.ndarray, 
               res: np.ndarray) -> float:
    """Computes Area density - approximate enclosing ellipsoid feature.
    This feature refers to "Fmorph_a_dens_aee" (ID = RDD2) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        vol (ndarray): 3D volume, NON-QUANTIZED, continous imaging intensity distribution.
        mask_int (ndarray): Intensity mask.
        mask_morph (ndarray): Morphological mask.
        res (ndarray): [a,b,c] vector specfying the resolution of the volume in mm.
                       XYZ resolution (world), or JIK resolution (intrinsic matlab).

    Returns:
        float: Value of the Area density - approximate enclosing ellipsoid feature.
    """
    vol, mask_int, mask_morph = padding(vol, mask_int, mask_morph)
    _, _, _, xyz_morph, faces, vertices, _ = get_variables(vol, mask_int, mask_morph, res)
    area = get_mesh_area(faces, vertices)
    [major, minor, least] = get_axis_lengths(xyz_morph)
    a = 2*np.sqrt(major)
    b = 2*np.sqrt(minor)
    c = 2*np.sqrt(least)
    a_aee = get_area_dens_approx(a, b, c, 20)
    a_dens_aee = area / a_aee

    return a_dens_aee  # Area density - approximate enclosing ellipsoid

def v_dens_mvee(vol: np.ndarray, 
                mask_int: np.ndarray, 
                mask_morph: np.ndarray, 
                res: np.ndarray) -> float:
    """Computes Volume density - minimum volume enclosing ellipsoid feature.
    Subsequent singular value decomposition of matrix A and and
    taking the inverse of the square root of the diagonal of the
    sigma matrix will produce respective semi-axis lengths.
    Subsequent singular value decomposition of matrix A and
    taking the inverse of the square root of the diagonal of the
    sigma matrix will produce respective semi-axis lengths.
    This feature refers to "Fmorph_v_dens_mvee" (ID = SWZ1) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        vol (ndarray): 3D volume, NON-QUANTIZED, continous imaging intensity distribution.
        mask_int (ndarray): Intensity mask.
        mask_morph (ndarray): Morphological mask.
        res (ndarray): [a,b,c] vector specfying the resolution of the volume in mm.
                       XYZ resolution (world), or JIK resolution (intrinsic matlab).

    Returns:
        float: Value of the Volume density - minimum volume enclosing ellipsoid feature.
    """
    vol, mask_int, mask_morph = padding(vol, mask_int, mask_morph)
    _, _, _, _, faces, vertices, conv_hull = get_variables(vol, mask_int, mask_morph, res)
    volume = get_mesh_volume(faces, vertices)
    p = np.stack((conv_hull.points[conv_hull.simplices[:, 0], 0],
                    conv_hull.points[conv_hull.simplices[:, 1], 1],
                    conv_hull.points[conv_hull.simplices[:, 2], 2]), axis=1)
    A, _ = min_vol_ellipse(np.transpose(p), 0.01)
    # New semi-axis lengths
    _, Q, _ = np.linalg.svd(A)
    a = 1/np.sqrt(Q[2])
    b = 1/np.sqrt(Q[1])
    c = 1/np.sqrt(Q[0])
    v_mvee = (4*np.pi*a*b*c) / 3
    v_dens_mvee = volume / v_mvee

    return v_dens_mvee  # Volume density - minimum volume enclosing ellipsoid

def a_dens_mvee(vol: np.ndarray, 
                mask_int: np.ndarray, 
                mask_morph: np.ndarray, 
                res: np.ndarray) -> float:
    """Computes Area density - minimum volume enclosing ellipsoid feature.
    Subsequent singular value decomposition of matrix A and and
    taking the inverse of the square root of the diagonal of the
    sigma matrix will produce respective semi-axis lengths.
    Subsequent singular value decomposition of matrix A and
    taking the inverse of the square root of the diagonal of the
    sigma matrix will produce respective semi-axis lengths.
    This feature refers to "Fmorph_a_dens_mvee" (ID = BRI8) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        vol (ndarray): 3D volume, NON-QUANTIZED, continous imaging intensity distribution.
        mask_int (ndarray): Intensity mask.
        mask_morph (ndarray): Morphological mask.
        res (ndarray): [a,b,c] vector specfying the resolution of the volume in mm.
                       XYZ resolution (world), or JIK resolution (intrinsic matlab).

    Returns:
        float: Value of the Area density - minimum volume enclosing ellipsoid feature.
    """
    vol, mask_int, mask_morph = padding(vol, mask_int, mask_morph)
    _, _, _, _, faces, vertices, conv_hull = get_variables(vol, mask_int, mask_morph, res)
    area = get_mesh_area(faces, vertices)
    p = np.stack((conv_hull.points[conv_hull.simplices[:, 0], 0],
                    conv_hull.points[conv_hull.simplices[:, 1], 1],
                    conv_hull.points[conv_hull.simplices[:, 2], 2]), axis=1)
    A, _ = min_vol_ellipse(np.transpose(p), 0.01)
    # New semi-axis lengths
    _, Q, _ = np.linalg.svd(A)
    a = 1/np.sqrt(Q[2])
    b = 1/np.sqrt(Q[1])
    c = 1/np.sqrt(Q[0])
    a_mvee = get_area_dens_approx(a, b, c, 20)
    a_dens_mvee = area / a_mvee

    return a_dens_mvee  # Area density - minimum volume enclosing ellipsoid

def v_dens_conv_hull(vol: np.ndarray, 
                     mask_int: np.ndarray, 
                     mask_morph: np.ndarray, 
                     res: np.ndarray) -> float:
    """Computes Volume density - convex hull feature.
    This feature refers to "Fmorph_v_dens_conv_hull" (ID = R3ER) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        vol (ndarray): 3D volume, NON-QUANTIZED, continous imaging intensity distribution.
        mask_int (ndarray): Intensity mask.
        mask_morph (ndarray): Morphological mask.
        res (ndarray): [a,b,c] vector specfying the resolution of the volume in mm.
                       XYZ resolution (world), or JIK resolution (intrinsic matlab).

    Returns:
        float: Value of the Volume density - convex hull feature.
    """
    vol, mask_int, mask_morph = padding(vol, mask_int, mask_morph)
    _, _, _, _, faces, vertices, conv_hull = get_variables(vol, mask_int, mask_morph, res)
    volume = get_mesh_volume(faces, vertices)
    v_convex = conv_hull.volume
    v_dens_conv_hull = volume / v_convex

    return v_dens_conv_hull  # Volume density - convex hull

def a_dens_conv_hull(vol: np.ndarray, 
                     mask_int: np.ndarray, 
                     mask_morph: np.ndarray, 
                     res: np.ndarray) -> float:
    """Computes Area density - convex hull feature.
    This feature refers to "Fmorph_a_dens_conv_hull" (ID = 7T7F) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        vol (ndarray): 3D volume, NON-QUANTIZED, continous imaging intensity distribution.
        mask_int (ndarray): Intensity mask.
        mask_morph (ndarray): Morphological mask.
        res (ndarray): [a,b,c] vector specfying the resolution of the volume in mm.
                       XYZ resolution (world), or JIK resolution (intrinsic matlab).

    Returns:
        float: Value of the Area density - convex hull feature.
    """
    vol, mask_int, mask_morph = padding(vol, mask_int, mask_morph)
    _, _, _, _, faces, vertices, conv_hull = get_variables(vol, mask_int, mask_morph, res)
    area = get_mesh_area(faces, vertices)
    v_convex = conv_hull.area
    a_dens_conv_hull = area / v_convex

    return a_dens_conv_hull  # Area density - convex hull

def integ_int(vol: np.ndarray, 
              mask_int: np.ndarray, 
              mask_morph: np.ndarray, 
              res: np.ndarray) -> float:
    """Computes Integrated intensity feature.
    This feature refers to "Fmorph_integ_int" (ID = 99N0) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        vol (ndarray): 3D volume, NON-QUANTIZED, continous imaging intensity distribution.
        mask_int (ndarray): Intensity mask.
        mask_morph (ndarray): Morphological mask.
        res (ndarray): [a,b,c] vector specfying the resolution of the volume in mm.
                       XYZ resolution (world), or JIK resolution (intrinsic matlab).

    Returns:
        float: Value of the Integrated intensity feature.

    """
    vol, mask_int, mask_morph = padding(vol, mask_int, mask_morph)
    xgl_int, _, _, _, faces, vertices, _ = get_variables(vol, mask_int, mask_morph, res)
    volume = get_mesh_volume(faces, vertices)
    integ_int = np.mean(xgl_int) * volume

    return integ_int  # Integrated intensity

def moran_i(vol: np.ndarray, 
            mask_int: np.ndarray, 
            mask_morph: np.ndarray, 
            res: np.ndarray, 
            compute_moran_i: bool=False) -> float:
    """Computes Moran's I index feature.
    This feature refers to "Fmorph_moran_i" (ID = N365) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        vol (ndarray): 3D volume, NON-QUANTIZED, continous imaging intensity distribution.
        mask_int (ndarray): Intensity mask.
        mask_morph (ndarray): Morphological mask.
        res (ndarray): [a,b,c] vector specfying the resolution of the volume in mm.
                       XYZ resolution (world), or JIK resolution (intrinsic matlab).
        compute_moran_i (bool, optional): True to compute Moran's Index.

    Returns:
        float: Value of the Moran's I index feature.

    """
    vol, mask_int, mask_morph = padding(vol, mask_int, mask_morph)

    if compute_moran_i:
        vol_mor = vol.copy()
        vol_mor[mask_int == 0] = np.NaN
        moran_i = get_moran_i(vol_mor, res)

    return moran_i  # Moran's I index

def geary_c(vol: np.ndarray, 
            mask_int: np.ndarray, 
            mask_morph: np.ndarray, 
            res: np.ndarray, 
            compute_geary_c: bool=False) -> float:
    """Computes Geary's C measure feature.
    This feature refers to "Fmorph_geary_c" (ID = NPT7) in 
    the `IBSI1 reference manual <https://arxiv.org/pdf/1612.07003.pdf>`__.

    Args:
        vol (ndarray): 3D volume, NON-QUANTIZED, continous imaging intensity distribution.
        mask_int (ndarray): Intensity mask.
        mask_morph (ndarray): Morphological mask.
        res (ndarray): [a,b,c] vector specfying the resolution of the volume in mm.
                       XYZ resolution (world), or JIK resolution (intrinsic matlab).
        compute_geary_c (bool, optional): True to compute Geary's C measure.

    Returns:
        float: Value of the Geary's C measure feature.

    """
    vol, mask_int, mask_morph = padding(vol, mask_int, mask_morph)

    if compute_geary_c:
        vol_mor = vol.copy()
        vol_mor[mask_int == 0] = np.NaN
        geary_c = get_geary_c(vol_mor, res)

    return geary_c  # Geary's C measure



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import logging
import random
from json import load
from pathlib import Path
from typing import Dict, List, Union

import numpy as np
import pandas as pd

from ..utils.get_patient_id_from_scan_name import get_patient_id_from_scan_name
from ..utils.initialize_features_names import initialize_features_names


def create_radiomics_table(radiomics_files_paths: List, image_space: str, log_file: Union[str, Path]) -> Dict:
    """
    Creates a dictionary with a csv and other information

    Args:
        radiomics_files_paths(List): List of paths to the radiomics JSON files.
        image_space(str): String of the image space that contains the extracted features
        log_file(Union[str, Path]): Path to logging file.

    Returns:
        Dict: Dictionary containing the extracted radiomics and other info (patientID, feature names...)
    """
    if log_file:
        # Setting up logging settings
        logging.basicConfig(filename=log_file, level=logging.DEBUG)
    
    # INITIALIZATIONS OF RADIOMICS STRUCTURES
    n_files = len(radiomics_files_paths)
    patientID = [0] * n_files
    rad_structs = [0] * n_files
    file_open = [False] * n_files

    for f in range(n_files):
        with open(radiomics_files_paths[f], "r") as fp: 
            radStruct = load(fp)
        rad_structs[f] = radStruct
        file_open[f] = True
        patientID[f] = get_patient_id_from_scan_name(radiomics_files_paths[f].stem)

    # INITIALIZE FEATURE NAMES
    logging.info(f"\nnFiles: {n_files}")
    non_text_cell = []
    text_cell = []
    while len(non_text_cell) == 0 and len(text_cell) == 0:
        try:
            rand_patient = np.floor(n_files * random.uniform(0, 1)).astype(int)
            with open(radiomics_files_paths[rand_patient], "r") as fp: 
                radiomics_struct = load(fp)

            # IMAGE SPACE STRUCTURE --> .morph, .locInt, ...,  .texture
            image_space_struct = radiomics_struct[image_space]
            non_text_cell, text_cell = initialize_features_names(image_space_struct)
        except:
            pass

    # CREATE TABLE DATA
    features_name_dict = {}
    str_table = ''
    str_names = '||'
    count_var = 0

    # Non-texture features
    for im_type in range(len(non_text_cell[0])):
        for param in range(len(non_text_cell[2][im_type])):
            for feat in range(len(non_text_cell[1][im_type])):
                count_var = count_var + 1
                feature_name = 'radVar' + str(count_var)
                features_name_dict.update({feature_name: [0] * n_files})
                real_name_feature = non_text_cell[0][im_type] + '__' + \
                    non_text_cell[1][im_type][feat] + '__' + \
                    non_text_cell[2][im_type][param]
                str_table = str_table + feature_name + ','
                str_names = str_names + feature_name + ':' + real_name_feature + '||'

                for f in range(n_files):
                    if file_open[f]:
                        try:
                            val = rad_structs[f][image_space][
                                non_text_cell[0][im_type]][
                                non_text_cell[2][im_type][param]][
                                non_text_cell[1][im_type][feat]]
                        except:
                            val = np.NaN
                        if type(val) in [str, list]:
                            val = np.NaN
                    else:
                        val = np.NaN
                    features_name_dict[feature_name][f] = val

    # Texture features
    for im_type in range(len(text_cell[0])):
        for param in range(len(text_cell[2][im_type])):
            for feat in range(len(text_cell[1][im_type])):
                count_var = count_var + 1
                feature_name = 'radVar' + str(count_var)
                features_name_dict.update({feature_name: [0] * n_files})
                real_name_feature = text_cell[0][im_type] + '__' + \
                    text_cell[1][im_type][feat] + '__' + \
                    text_cell[2][im_type][param]
                str_table = str_table + feature_name + ','
                str_names = str_names + feature_name + ':' + real_name_feature + '||'
                for f in range(n_files):
                    if file_open[f]:
                        try:
                            val = rad_structs[f][image_space]['texture'][
                                text_cell[0][im_type]][
                                text_cell[2][im_type][param]][
                                text_cell[1][im_type][feat]]
                        except:
                            val = np.NaN
                        if type(val) in [str, list]:
                            val = np.NaN
                    else:
                        val = np.NaN
                    features_name_dict[feature_name][f] = val

    radiomics_table_dict = {
        'Table': pd.DataFrame(features_name_dict, index=patientID),
        'Properties': {'UserData': str_names,
                       'RowNames': patientID,
                       'DimensionNames': ['PatientID', 'Variables'],
                       'VariableNames': [key for key in features_name_dict.keys()]
                       }}

    return radiomics_table_dict



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import math
from typing import List, Tuple, Union

import numpy as np
from skimage.measure import marching_cubes


def find_i_x(levels: np.ndarray,
             fract_vol: np.ndarray,
             x: float) -> np.ndarray:
    """Computes intensity at volume fraction.

    Args:
        levels (ndarray): COMPLETE INTEGER grey-levels.
        fract_vol (ndarray): Fractional volume.
        x (float): Fraction percentage, between 0 and 100.

    Returns:
        ndarray: Array of minimum discretised intensity present in at most :math:`x` % of the volume.
    
    """
    ind = np.where(fract_vol <= x/100)[0][0]
    ix = levels[ind]

    return ix
    
def find_v_x(fract_int: np.ndarray,
             fract_vol: np.ndarray,
             x: float) -> np.ndarray:
    """Computes volume at intensity fraction.

    Args:
        fract_int (ndarray): Intensity fraction.
        fract_vol (ndarray): Fractional volume.
        x (float): Fraction percentage, between 0 and 100.

    Returns:
        ndarray: Array of largest volume fraction ``fract_vol`` that has an
        intensity fraction ``fract_int`` of at least :math:`x` %.

    """
    ind = np.where(fract_int >= x/100)[0][0]
    vx = fract_vol[ind]

    return vx

def get_area_dens_approx(a: float,
                         b: float,
                         c: float,
                         n: float) -> float:
    """Computes area density - minimum volume enclosing ellipsoid
    
    Args:
        a (float): Major semi-axis length.
        b (float): Minor semi-axis length.
        c (float): Least semi-axis length.
        n (int): Number of iterations.

    Returns:
        float: Area density - minimum volume enclosing ellipsoid.

    """
    alpha = np.sqrt(1 - b**2/a**2)
    beta = np.sqrt(1 - c**2/a**2)
    ab = alpha * beta
    point = (alpha**2+beta**2) / (2*ab)
    a_ell = 0

    for v in range(0, n+1):
        coef = [0]*v + [1]
        legen = np.polynomial.legendre.legval(x=point, c=coef)
        a_ell = a_ell + ab**v / (1-4*v**2) * legen

    a_ell = a_ell * 4 * np.pi * a * b

    return a_ell

def get_axis_lengths(xyz: np.ndarray) -> Tuple[float, float, float]:
    """Computes AxisLengths.
    
    Args:
        xyz (ndarray): Array of three column vectors, defining the [X,Y,Z]
            positions of the points in the ROI (1's) of the mask volume. In mm.

    Returns:
        Tuple[float, float, float]: Array of three column vectors 
            [Major axis lengths, Minor axis lengths, Least axis lengths].

    """
    xyz = xyz.copy()

    # Getting the geometric centre of mass
    com_geom = np.sum(xyz, 0)/np.shape(xyz)[0]  # [1 X 3] vector

    # Subtracting the centre of mass
    xyz[:, 0] = xyz[:, 0] - com_geom[0]
    xyz[:, 1] = xyz[:, 1] - com_geom[1]
    xyz[:, 2] = xyz[:, 2] - com_geom[2]

    # Getting the covariance matrix
    cov_mat = np.cov(xyz, rowvar=False)

    # Getting the eigenvalues
    eig_val, _ = np.linalg.eig(cov_mat)
    eig_val = np.sort(eig_val)

    major = eig_val[2]
    minor = eig_val[1]
    least = eig_val[0]

    return major, minor, least

def get_glcm_cross_diag_prob(p_ij: np.ndarray) -> np.ndarray:
    """Computes cross diagonal probabilities.

    Args:
        p_ij (ndarray): Joint probability of grey levels 
            i and j occurring in neighboring voxels. (Elements
            of the  probability distribution for grey level 
            co-occurrences).

    Returns:
        ndarray: Array of the cross diagonal probability.
     
    """
    n_g = np.size(p_ij, 0)
    val_k = np.arange(2, 2*n_g + 100*np.finfo(float).eps)
    n_k = np.size(val_k)
    p_iplusj = np.zeros(n_k)

    for iteration_k in range(0, n_k):
        k = val_k[iteration_k]
        p = 0
        for i in range(0, n_g):
            for j in range(0, n_g):
                if (k - (i+j+2)) == 0:
                    p += p_ij[i, j]

        p_iplusj[iteration_k] = p

    return p_iplusj

def get_glcm_diag_prob(p_ij: np.ndarray) -> np.ndarray:
    """Computes diagonal probabilities.

    Args:
        p_ij (ndarray): Joint probability of grey levels 
            i and j occurring in neighboring voxels. (Elements
            of the  probability distribution for grey level 
            co-occurrences).

    Returns:
        ndarray: Array of the diagonal probability.
    
    """

    n_g = np.size(p_ij, 0)
    val_k = np.arange(0, n_g)
    n_k = np.size(val_k)
    p_iminusj = np.zeros(n_k)

    for iteration_k in range(0, n_k):
        k = val_k[iteration_k]
        p = 0
        for i in range(0, n_g):
            for j in range(0, n_g):
                if (k - abs(i-j)) == 0:
                    p += p_ij[i, j]

        p_iminusj[iteration_k] = p

    return p_iminusj

def get_com(xgl_int: np.ndarray,
            xgl_morph: np.ndarray,
            xyz_int: np.ndarray,
            xyz_morph: np.ndarray) -> Union[float,
                                            np.ndarray]:
    """Calculates center of mass shift (in mm, since resolution is in mm).

    Note: 
        Row positions of "x_gl" and "xyz" must correspond for each point.
    
    Args:
        xgl_int (ndarray): Vector of intensity values in the volume to analyze 
            (only values in the intensity mask).
        xgl_morph (ndarray): Vector of intensity values in the volume to analyze 
            (only values in the morphological mask).
        xyz_int (ndarray): [n_points X 3] matrix of three column vectors, defining the [X,Y,Z]
            positions of the points in the ROI (1's) of the mask volume (In mm).
            (Mesh-based volume calculated from the ROI intensity mesh)
        xyz_morph (ndarray): [n_points X 3] matrix of three column vectors, defining the [X,Y,Z]
            positions of the points in the ROI (1's) of the mask volume (In mm).
            (Mesh-based volume calculated from the ROI morphological mesh)

    Returns:
        Union[float, np.ndarray]: The ROI volume centre of mass.

    """

    # Getting the geometric centre of mass
    n_v = np.size(xgl_morph)

    com_geom = np.sum(xyz_morph, 0)/n_v  # [1 X 3] vector

    # Getting the density centre of mass
    xyz_int[:, 0] = xgl_int*xyz_int[:, 0]
    xyz_int[:, 1] = xgl_int*xyz_int[:, 1]
    xyz_int[:, 2] = xgl_int*xyz_int[:, 2]
    com_gl = np.sum(xyz_int, 0)/np.sum(xgl_int, 0)  # [1 X 3] vector

    # Calculating the shift
    com = np.linalg.norm(com_geom - com_gl)

    return com

def get_loc_peak(img_obj: np.ndarray,
                 roi_obj: np.ndarray,
                 res: np.ndarray) -> float:
    """Computes Local intensity peak.

    Note:
        This works only in 3D for now.

    Args:
        img_obj (ndarray): Continuos image intensity distribution, with no NaNs
            outside the ROI.
        roi_obj (ndarray): Array of the mask defining the ROI.
        res (List[float]): [a,b,c] vector specifying the resolution of the volume in mm.
            xyz resolution (world), or JIK resolution (intrinsic matlab).

    Returns:
        float: Value of the local intensity peak.
    
    """
    # INITIALIZATION
    # About 6.2 mm, as defined in document
    dist_thresh = (3/(4*math.pi))**(1/3)*10

    # Insert -inf outside ROI
    temp = img_obj.copy()
    img_obj = img_obj.copy()
    img_obj[roi_obj == 0] = -np.inf

    # Find the location(s) of the maximal voxel
    max_val = np.max(img_obj)
    I, J, K = np.nonzero(img_obj == max_val)
    n_max = np.size(I)

    # Reconverting to full object without -Inf
    img_obj = temp

    # Get a meshgrid first
    x = res[0]*(np.arange(img_obj.shape[1])+0.5)
    y = res[1]*(np.arange(img_obj.shape[0])+0.5)
    z = res[2]*(np.arange(img_obj.shape[2])+0.5)
    X, Y, Z = np.meshgrid(x, y, z)  # In mm

    # Calculate the local peak
    max_val = -np.inf

    for n in range(n_max):
        temp_x = X - X[I[n], J[n], K[n]]
        temp_y = Y - Y[I[n], J[n], K[n]]
        temp_z = Z - Z[I[n], J[n], K[n]]
        temp_dist_mesh = (np.sqrt(np.power(temp_x, 2) + 
                                np.power(temp_y, 2) +
                                np.power(temp_z, 2)))
        val = img_obj[temp_dist_mesh <= dist_thresh]
        val[np.isnan(val)] = []

        if np.size(val) == 0:
            temp_local_peak = img_obj[I[n], J[n], K[n]]
        else:
            temp_local_peak = np.mean(val)
        if temp_local_peak > max_val:
            max_val = temp_local_peak

    local_peak = max_val

    return local_peak

def get_mesh(mask: np.ndarray,
             res: Union[np.ndarray, List]) -> Tuple[np.ndarray,
                                                    np.ndarray,
                                                    np.ndarray]:
    """Compute Mesh.

    Note:
      Make sure the `mask` is padded with a layer of 0's in all
      dimensions to reduce potential isosurface computation errors.

    Args:
        mask (ndarray): Contains only 0's and 1's.
        res (ndarray or List): [a,b,c] vector specifying the resolution of the volume in mm.
            xyz resolution (world), or JIK resolution (intrinsic matlab).

    Returns:
        Tuple[np.ndarray, np.ndarray, np.ndarray]: 
            - Array of the [X,Y,Z] positions of the ROI.
            - Array of the spatial coordinates for `mask` unique mesh vertices.
            - Array of triangular faces via referencing vertex indices from vertices.
    """
    # Getting the grid of X,Y,Z positions, where the coordinate reference
    # system (0,0,0) is located at the upper left corner of the first voxel
    # (-0.5: half a voxel distance). For the whole volume defining the mask,
    # no matter if it is a 1 or a 0.
    mask = mask.copy()
    res = res.copy()

    x = res[0]*((np.arange(1, np.shape(mask)[0]+1))-0.5)
    y = res[1]*((np.arange(1, np.shape(mask)[1]+1))-0.5)
    z = res[2]*((np.arange(1, np.shape(mask)[2]+1))-0.5)
    X, Y, Z = np.meshgrid(x, y, z, indexing='ij')

    # Getting the isosurface of the mask
    vertices, faces, _, _ = marching_cubes(volume=mask, level=0.5, spacing=res)

    # Getting the X,Y,Z positions of the ROI (i.e. 1's) of the mask
    X = np.reshape(X, (np.size(X), 1), order='F')
    Y = np.reshape(Y, (np.size(Y), 1), order='F')
    Z = np.reshape(Z, (np.size(Z), 1), order='F')

    xyz = np.concatenate((X, Y, Z), axis=1)
    xyz = xyz[np.where(np.reshape(mask, np.size(mask), order='F') == 1)[0], :]

    return xyz, faces, vertices

def get_glob_peak(img_obj: np.ndarray,
                  roi_obj: np.ndarray,
                  res: np.ndarray) -> float:
    """Computes Global intensity peak.

    Note:
        This works only in 3D for now.

    Args:
        img_obj (ndarray): Continuos image intensity distribution, with no NaNs
            outside the ROI.
        roi_obj (ndarray): Array of the mask defining the ROI.
        res (List[float]): [a,b,c] vector specifying the resolution of the volume in mm.
            xyz resolution (world), or JIK resolution (intrinsic matlab).

    Returns:
        float: Value of the global intensity peak.

    """
    # INITIALIZATION
    # About 6.2 mm, as defined in document
    dist_thresh = (3/(4*math.pi))**(1/3)*10

    # Find the location(s) of all voxels within the ROI
    indices = np.nonzero(np.reshape(roi_obj, np.size(roi_obj), order='F') == 1)[0]
    I, J, K = np.unravel_index(indices, np.shape(img_obj), order='F')
    n_max = np.size(I)

    # Get a meshgrid first
    x = res[0]*(np.arange(img_obj.shape[1])+0.5)
    y = res[1]*(np.arange(img_obj.shape[0])+0.5)
    z = res[2]*(np.arange(img_obj.shape[2])+0.5)
    X, Y, Z = np.meshgrid(x, y, z)  # In mm

    # Calculate the local peak
    max_val = -np.inf

    for n in range(n_max):
        temp_x = X - X[I[n], J[n], K[n]]
        temp_y = Y - Y[I[n], J[n], K[n]]
        temp_z = Z - Z[I[n], J[n], K[n]]
        temp_dist_mesh = (np.sqrt(np.power(temp_x, 2) + 
                                np.power(temp_y, 2) +
                                np.power(temp_z, 2)))
        val = img_obj[temp_dist_mesh <= dist_thresh]
        val[np.isnan(val)] = []

        if np.size(val) == 0:
            temp_local_peak = img_obj[I[n], J[n], K[n]]
        else:
            temp_local_peak = np.mean(val)
        if temp_local_peak > max_val:
            max_val = temp_local_peak

    global_peak = max_val

    return global_peak
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-


def get_scan_name_from_rad_name(rad_name: str) -> str:
    """Finds the imaging scan name from thr radiomics structure name

    Args:
        rad_name (str): radiomics structure name.
    
    Returns:
        str: String of the imaging scan name
    
    Example:
        >>> get_scan_name_from_rad_name('STS-McGill-001__T1(tumourAndEdema).MRscan')
        'T1'
    """
    ind_double_under = rad_name.find('__')
    ind_open_par = rad_name.find('(')
    scan_name = rad_name[ind_double_under + 2:ind_open_par]

    return scan_name


#!/usr/bin/env python3
# -*- coding: utf-8 -*-


from pathlib import Path
from typing import Dict, Union
import SimpleITK as sitk
import numpy as np


def image_reader_SITK(path: Path, 
                      option: str=None) -> Union[Dict, None]:
    """Return the image in a numpy array or a dictionary with the header of the image.

    Args:
        path (path): path of the file
        option (str): name of the option, either 'image' or 'header'
    
    Returns:
        Union[Dict, None]: dictionary with the header of the image
    """
    if option is None or option == 'image':
        # return the image in a numpy array
        return np.transpose(sitk.GetArrayFromImage(sitk.ReadImage(path)))
    elif option == 'header':
        # Return a dictionary with the header of the image.
        reader = sitk.ImageFileReader()
        reader.SetFileName(path)
        # reader.LoadPrivateTagsOn()
        reader.ReadImageInformation()
        dic_im_header = {}
        for key in reader.GetMetaDataKeys():
            dic_im_header.update({key: reader.GetMetaData(key)})
        return dic_im_header
    else:
        print("Argument option should be the string 'image' or 'header'")
        return None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import sys
import warnings
from typing import List, Union

import numpy as np
import pydicom
import ray

from ..utils.imref import imref3d

warnings.simplefilter("ignore")

from pathlib import Path

from ..MEDscan import MEDscan
from ..processing.segmentation import get_roi
from ..utils.save_MEDscan import save_MEDscan


class ProcessDICOM():
    """
    Class to process dicom files and extract imaging volume and 3D masks from it
    in order to oganize the data in a MEDscan class object.
    """

    def __init__(
        self,
        path_images: List[Path],
        path_rs: List[Path],
        path_save: Union[str, Path],
        save: bool) -> None:
        """
        Args:
            path_images (List[Path]): List of paths to the dicom files of a single scan.
            path_rs (List[Path]): List of paths to the RT struct dicom files for the same scan.
            path_save (Union[str, Path]): Path to the folder where the MEDscan object will be saved.
            save (bool): Whether to save the MEDscan object or not.
        
        Returns:
            None.
        """
        self.path_images = path_images
        self.path_rs = path_rs
        self.path_save = Path(path_save) if path_save is str else path_save
        self.save = save
    
    def __get_dicom_scan_orientation(self, dicom_header: List[pydicom.dataset.FileDataset]) -> str:
        """
        Get the orientation of the scan.
        
        Args:
            dicom_header (List[pydicom.dataset.FileDataset]): List of dicom headers.
        
        Returns:
            str: Orientation of the scan.
        """
        n_slices = len(dicom_header)
        image_patient_positions_x = [dicom_header[i].ImagePositionPatient[0] for i in range(n_slices)]
        image_patient_positions_y = [dicom_header[i].ImagePositionPatient[1] for i in range(n_slices)]
        image_patient_positions_z = [dicom_header[i].ImagePositionPatient[2] for i in range(n_slices)]
        dist = [
            np.median(np.abs(np.diff(image_patient_positions_x))), 
            np.median(np.abs(np.diff(image_patient_positions_y))), 
            np.median(np.abs(np.diff(image_patient_positions_z)))
        ]
        index = dist.index(max(dist))
        if index == 0:
            orientation = 'Sagittal'
        elif index == 1:
            orientation = 'Coronal'
        else:
            orientation = 'Axial'
        
        return orientation

    def __merge_slice_pixel_arrays(self, slice_datasets):
        first_dataset = slice_datasets[0]
        num_rows = first_dataset.Rows
        num_columns = first_dataset.Columns
        num_slices = len(slice_datasets)

        sorted_slice_datasets = self.__sort_by_slice_spacing(slice_datasets)

        if any(self.__requires_rescaling(d) for d in sorted_slice_datasets):
            voxels = np.empty(
                (num_columns, num_rows, num_slices), dtype=np.float32)
            for k, dataset in enumerate(sorted_slice_datasets):
                slope = float(getattr(dataset, 'RescaleSlope', 1))
                intercept = float(getattr(dataset, 'RescaleIntercept', 0))
                voxels[:, :, k] = dataset.pixel_array.T.astype(
                    np.float32)*slope + intercept
        else:
            dtype = first_dataset.pixel_array.dtype
            voxels = np.empty((num_columns, num_rows, num_slices), dtype=dtype)
            for k, dataset in enumerate(sorted_slice_datasets):
                voxels[:, :, k] = dataset.pixel_array.T

        return voxels

    def __requires_rescaling(self, dataset):
        return hasattr(dataset, 'RescaleSlope') or hasattr(dataset, 'RescaleIntercept')

    def __ijk_to_patient_xyz_transform_matrix(self, slice_datasets):
        first_dataset = self.__sort_by_slice_spacing(slice_datasets)[0]
        image_orientation = first_dataset.ImageOrientationPatient
        row_cosine, column_cosine, slice_cosine = self.__extract_cosines(
            image_orientation)

        row_spacing, column_spacing = first_dataset.PixelSpacing
        slice_spacing = self.__slice_spacing(slice_datasets)

        transform = np.identity(4, dtype=np.float32)
        rotation = np.identity(3, dtype=np.float32)
        scaling = np.identity(3, dtype=np.float32)

        transform[:3, 0] = row_cosine*column_spacing
        transform[:3, 1] = column_cosine*row_spacing
        transform[:3, 2] = slice_cosine*slice_spacing

        transform[:3, 3] = first_dataset.ImagePositionPatient

        rotation[:3, 0] = row_cosine
        rotation[:3, 1] = column_cosine
        rotation[:3, 2] = slice_cosine

        rotation = np.transpose(rotation)

        scaling[0, 0] = column_spacing
        scaling[1, 1] = row_spacing
        scaling[2, 2] = slice_spacing

        return transform, rotation, scaling

    def __validate_slices_form_uniform_grid(self, slice_datasets):
        """
        Perform various data checks to ensure that the list of slices form a
        evenly-spaced grid of data.
        Some of these checks are probably not required if the data follows the
        DICOM specification, however it seems pertinent to check anyway.
        """
        invariant_properties = [
            'Modality',
            'SOPClassUID',
            'SeriesInstanceUID',
            'Rows',
            'Columns',
            'ImageOrientationPatient',
            'PixelSpacing',
            'PixelRepresentation',
            'BitsAllocated',
            'BitsStored',
            'HighBit',
        ]

        for property_name in invariant_properties:
            self.__slice_attribute_equal(slice_datasets, property_name)

        self.__validate_image_orientation(slice_datasets[0].ImageOrientationPatient)

        slice_positions = self.__slice_positions(slice_datasets)
        self.__check_for_missing_slices(slice_positions)

    def __validate_image_orientation(self, image_orientation):
        """
        Ensure that the image orientation is supported
        - The direction cosines have magnitudes of 1 (just in case)
        - The direction cosines are perpendicular
        """

        row_cosine, column_cosine, slice_cosine = self.__extract_cosines(
            image_orientation)

        if not self.__almost_zero(np.dot(row_cosine, column_cosine), 1e-4):
            raise ValueError(
                "Non-orthogonal direction cosines: {}, {}".format(row_cosine, column_cosine))
        elif not self.__almost_zero(np.dot(row_cosine, column_cosine), 1e-8):
            warnings.warn("Direction cosines aren't quite orthogonal: {}, {}".format(
                row_cosine, column_cosine))

        if not self.__almost_one(np.linalg.norm(row_cosine), 1e-4):
            raise ValueError(
                "The row direction cosine's magnitude is not 1: {}".format(row_cosine))
        elif not self.__almost_one(np.linalg.norm(row_cosine), 1e-8):
            warnings.warn(
                "The row direction cosine's magnitude is not quite 1: {}".format(row_cosine))

        if not self.__almost_one(np.linalg.norm(column_cosine), 1e-4):
            raise ValueError(
                "The column direction cosine's magnitude is not 1: {}".format(column_cosine))
        elif not self.__almost_one(np.linalg.norm(column_cosine), 1e-8):
            warnings.warn(
                "The column direction cosine's magnitude is not quite 1: {}".format(column_cosine))
        sys.stderr.flush()

    def __is_close(self, a, b, rel_tol=1e-9, abs_tol=0.0):
        return abs(a-b) <= max(rel_tol*max(abs(a), abs(b)), abs_tol)

    def __almost_zero(self, value, abs_tol):
        return self.__is_close(value, 0.0, abs_tol=abs_tol)

    def __almost_one(self, value, abs_tol):
        return self.__is_close(value, 1.0, abs_tol=abs_tol)

    def __extract_cosines(self, image_orientation):
        row_cosine = np.array(image_orientation[:3])
        column_cosine = np.array(image_orientation[3:])
        slice_cosine = np.cross(row_cosine, column_cosine)
        return row_cosine, column_cosine, slice_cosine

    def __slice_attribute_equal(self, slice_datasets, property_name):
        initial_value = getattr(slice_datasets[0], property_name, None)
        for slice_idx, dataset in enumerate(slice_datasets[1:]):
            value = getattr(dataset, property_name, None)
            if value != initial_value:
                msg = f'Slice {slice_idx+1} have different value for {property_name}: {value} != {initial_value}'
                warnings.warn(msg)

    def __slice_positions(self, slice_datasets):
        image_orientation = slice_datasets[0].ImageOrientationPatient
        row_cosine, column_cosine, slice_cosine = self.__extract_cosines(
            image_orientation)
        return [np.dot(slice_cosine, d.ImagePositionPatient) for d in slice_datasets]

    def __check_for_missing_slices(self, slice_positions):
        slice_positions_diffs = np.diff(sorted(slice_positions))
        if not np.allclose(slice_positions_diffs, slice_positions_diffs[0], atol=0, rtol=1e-5):
            msg = "The slice spacing is non-uniform. Slice spacings:\n{}"
            warnings.warn(msg.format(slice_positions_diffs))
            sys.stderr.flush()
        if not np.allclose(slice_positions_diffs, slice_positions_diffs[0], atol=0, rtol=1e-1):
            raise ValueError('The slice spacing is non-uniform. It appears there are extra slices from another scan')

    def __slice_spacing(self, slice_datasets):
        if len(slice_datasets) > 1:
            slice_positions = self.__slice_positions(slice_datasets)
            slice_positions_diffs = np.diff(sorted(slice_positions))
            return np.mean(slice_positions_diffs)

        return 0.0

    def __sort_by_slice_spacing(self, slice_datasets):
        slice_spacing = self.__slice_positions(slice_datasets)
        return [d for (s, d) in sorted(zip(slice_spacing, slice_datasets))]

    def combine_slices(self, slice_datasets: List[pydicom.dataset.FileDataset]) -> List[np.ndarray]:
        """
        Given a list of pydicom datasets for an image series, stitch them together into a
        three-dimensional numpy array of iamging data. Also calculate a 4x4 affine transformation
        matrix that converts the ijk-pixel-indices into the xyz-coordinates in the
        DICOM patient's coordinate system and 4x4 rotation and scaling matrix.
        If any of the DICOM images contain either the
        `Rescale Slope <https://dicom.innolitics.com/ciods/ct-image/ct-image/00281053>`__ or the
        `Rescale Intercept <https://dicom.innolitics.com/ciods/ct-image/ct-image/00281052>`__
        attributes they will be applied to each slice individually.
        This function requires that the datasets:

        - Be in same series (have the same
          `Series Instance UID <https://dicom.innolitics.com/ciods/ct-image/general-series/0020000e>`__,
          `Modality <https://dicom.innolitics.com/ciods/ct-image/general-series/00080060>`__,
          and `SOP Class UID <https://dicom.innolitics.com/ciods/ct-image/sop-common/00080016>`__).
        - The binary storage of each slice must be the same (have the same
          `Bits Allocated <https://dicom.innolitics.com/ciods/ct-image/image-pixel/00280100>`__,
          `Bits Stored <https://dicom.innolitics.com/ciods/ct-image/image-pixel/00280101>`__,
          `High Bit <https://dicom.innolitics.com/ciods/ct-image/image-pixel/00280102>`__, and
          `Pixel Representation <https://dicom.innolitics.com/ciods/ct-image/image-pixel/00280103>`__).
        - The image slice must approximately form a grid. This means there can not
          be any missing internal slices (missing slices on the ends of the dataset
          are not detected). It also means that  each slice must have the same
          `Rows <https://dicom.innolitics.com/ciods/ct-image/image-pixel/00280010>`__,
          `Columns <https://dicom.innolitics.com/ciods/ct-image/image-pixel/00280011>`__,
          `Pixel Spacing <https://dicom.innolitics.com/ciods/ct-image/image-plane/00280030>`__, and
          `Image Orientation (Patient) <https://dicom.innolitics.com/ciods/ct-image/image-plane/00200037>`__
          attribute values.
        - The direction cosines derived from the
          `Image Orientation (Patient) <https://dicom.innolitics.com/ciods/ct-image/image-plane/00200037>`__
          attribute must, within 1e-4, have a magnitude of 1.  The cosines must
          also be approximately perpendicular (their dot-product must be within
          1e-4 of 0).  Warnings are displayed if any of theseapproximations are
          below 1e-8, however, since we have seen real datasets with values up to
          1e-4, we let them pass.
        - The `Image Position (Patient) <https://dicom.innolitics.com/ciods/ct-image/image-plane/00200032>`__
          values must approximately form a line.
        
        If any of these conditions are not met, a `dicom_numpy.DicomImportException` is raised.

        Args:
            slice_datasets (List[pydicom.dataset.FileDataset]): List of dicom headers.
        Returns:
            List[numpy.ndarray]: List of numpy arrays containing the data extracted the dicom files
            (voxels, translation, rotation and scaling matrix).
        """

        if not slice_datasets:
            raise ValueError("Must provide at least one DICOM dataset")

        self.__validate_slices_form_uniform_grid(slice_datasets)

        voxels = self.__merge_slice_pixel_arrays(slice_datasets)
        transform, rotation, scaling = self.__ijk_to_patient_xyz_transform_matrix(
            slice_datasets)

        return voxels, transform, rotation, scaling

    def process_files(self):
        """
        Reads DICOM files (imaging volume + ROIs) in the instance data path 
        and then organizes it in the MEDscan class.
        
        Args:
            None.
        
        Returns:
            medscan (MEDscan): Instance of a MEDscan class.
        """
        
        return self.process_files_wrapper.remote(self)
    
    @ray.remote
    def process_files_wrapper(self) -> MEDscan:
        """
        Wrapper function to process the files.
        """

        # PARTIAL PARSING OF ARGUMENTS
        if self.path_images is None:
            raise ValueError('At least two arguments must be provided')

        # INITIALIZATION
        medscan = MEDscan()

        # IMAGING DATA AND ROI DEFINITION (if applicable)
        # Reading DICOM images and headers
        dicom_hi = [pydicom.dcmread(str(dicom_file), force=True)
                for dicom_file in self.path_images]

        try:
            # Determination of the scan orientation
            medscan.data.orientation = self.__get_dicom_scan_orientation(dicom_hi)

            # IMPORTANT NOTE: extract_voxel_data using combine_slices from dicom_numpy
            # missing slices and oblique restrictions apply see the reference:
            # https://dicom-numpy.readthedocs.io/en/latest/index.html#dicom_numpy.combine_slices
            try:
                voxel_ndarray, ijk_to_xyz, rotation_m, scaling_m = self.combine_slices(dicom_hi)
            except ValueError as e:
                raise ValueError(f'Invalid DICOM data for combine_slices(). Error: {e}')

            # Alignment of scan coordinates for MR scans
            # (inverse of ImageOrientationPatient rotation matrix)
            if not np.allclose(rotation_m, np.eye(rotation_m.shape[0])):
                medscan.data.volume.scan_rot = rotation_m

            medscan.data.volume.array = voxel_ndarray
            medscan.type = dicom_hi[0].Modality + 'scan'

            # 7. Creation of imref3d object
            pixel_x = scaling_m[0, 0]
            pixel_y = scaling_m[1, 1]
            slice_s = scaling_m[2, 2]
            min_grid = rotation_m@ijk_to_xyz[:3, 3]
            min_x_grid = min_grid[0]
            min_y_grid = min_grid[1]
            min_z_grid = min_grid[2]
            size_image = np.shape(voxel_ndarray)
            spatial_ref = imref3d(size_image, pixel_x, pixel_y, slice_s)
            spatial_ref.XWorldLimits = (np.array(spatial_ref.XWorldLimits) -
                                    (spatial_ref.XWorldLimits[0] -
                                        (min_x_grid-pixel_x/2))).tolist()
            spatial_ref.YWorldLimits = (np.array(spatial_ref.YWorldLimits) -
                                    (spatial_ref.YWorldLimits[0] -
                                        (min_y_grid-pixel_y/2))).tolist()
            spatial_ref.ZWorldLimits = (np.array(spatial_ref.ZWorldLimits) -
                                    (spatial_ref.ZWorldLimits[0] -
                                        (min_z_grid-slice_s/2))).tolist()

            # Converting the results into lists
            spatial_ref.ImageSize = spatial_ref.ImageSize.tolist()
            spatial_ref.XIntrinsicLimits = spatial_ref.XIntrinsicLimits.tolist()
            spatial_ref.YIntrinsicLimits = spatial_ref.YIntrinsicLimits.tolist()
            spatial_ref.ZIntrinsicLimits = spatial_ref.ZIntrinsicLimits.tolist()

            # Update the spatial reference in the MEDscan class
            medscan.data.volume.spatialRef = spatial_ref
            
            # DICOM HEADERS OF IMAGING DATA
            dicom_h = [
                pydicom.dcmread(str(dicom_file),stop_before_pixels=True,force=True) for dicom_file in self.path_images
                ]
            for i in range(0, len(dicom_h)):
                dicom_h[i].remove_private_tags()
            medscan.dicomH = dicom_h

            # DICOM RTstruct (if applicable)
            if self.path_rs is not None and len(self.path_rs) > 0:
                dicom_rs_full = [
                    pydicom.dcmread(str(dicom_file),
                                    stop_before_pixels=True,
                                    force=True)
                    for dicom_file in self.path_rs
                ]
                for i in range(0, len(dicom_rs_full)):
                    dicom_rs_full[i].remove_private_tags()

            # GATHER XYZ POINTS OF ROIs USING RTstruct
            n_rs = len(dicom_rs_full) if type(dicom_rs_full) is list else dicom_rs_full
            contour_num = 0
            for rs in range(n_rs):
                n_roi = len(dicom_rs_full[rs].StructureSetROISequence)
                for roi in range(n_roi):
                    if roi!=0:
                        if dicom_rs_full[rs].StructureSetROISequence[roi].ROIName == \
                                dicom_rs_full[rs].StructureSetROISequence[roi-1].ROIName:
                            continue
                    points = []
                    name_set_strings = ['StructureSetName', 'StructureSetDescription',
                                    'series_description', 'SeriesInstanceUID']
                    for name_field in name_set_strings:
                        if name_field in dicom_rs_full[rs]:
                            name_set = getattr(dicom_rs_full[rs], name_field)
                            name_set_info = name_field
                            break

                    medscan.data.ROI.update_roi_name(key=contour_num,
                                                    roi_name=dicom_rs_full[rs].StructureSetROISequence[roi].ROIName)
                    medscan.data.ROI.update_indexes(key=contour_num,
                                                    indexes=None)
                    medscan.data.ROI.update_name_set(key=contour_num,
                                                    name_set=name_set)
                    medscan.data.ROI.update_name_set_info(key=contour_num,
                                                    nameSetInfo=name_set_info)
                    
                    try:
                        n_closed_contour = len(dicom_rs_full[rs].ROIContourSequence[roi].ContourSequence)
                        ind_closed_contour = []
                        for s in range(0, n_closed_contour):
                            # points stored in the RTstruct file for a given closed
                            # contour (beware: there can be multiple closed contours
                            # on a given slice).
                            pts_temp = dicom_rs_full[rs].ROIContourSequence[roi].ContourSequence[s].ContourData
                            n_points = int(len(pts_temp) / 3)
                            if len(pts_temp) > 0:
                                ind_closed_contour = ind_closed_contour + np.tile(s, n_points).tolist()
                                if type(points) == list:
                                    points = np.reshape(np.transpose(pts_temp),(n_points, 3))
                                else:
                                    points = np.concatenate(
                                            (points, np.reshape(np.transpose(pts_temp), (n_points, 3))),
                                            axis=0
                                            )
                        if n_closed_contour == 0:
                            print(f'Warning: no contour data found for ROI: \
                                  {dicom_rs_full[rs].StructureSetROISequence[roi].ROIName}')
                        else:
                            # Save the XYZ points in the MEDscan class
                            medscan.data.ROI.update_indexes(
                                                        key=contour_num, 
                                                        indexes=np.concatenate(
                                                                (points, 
                                                                np.reshape(ind_closed_contour, (len(ind_closed_contour), 1))),
                                                        axis=1)
                                                        )
                            # Compute the ROI box
                            _, roi_obj = get_roi(
                                            medscan,
                                            name_roi='{' + dicom_rs_full[rs].StructureSetROISequence[roi].ROIName + '}',
                                            box_string='full'
                                            )

                            # Save the ROI box non-zero indexes in the MEDscan class
                            medscan.data.ROI.update_indexes(key=contour_num, indexes=np.nonzero(roi_obj.data.flatten()))

                    except Exception as e:
                        if 'SeriesDescription' in dicom_h[0]:
                            print(f'patientID: {dicom_hi[0].PatientID} Modality: {dicom_hi[0].SeriesDescription} error: \
                                {str(e)} n_roi: {str(roi)}  n_rs: {str(rs)}')
                        else:
                            print(f'patientID: {dicom_hi[0].PatientID} Modality: {dicom_hi[0].Modality} error: \
                            {str(e)} n_roi: {str(roi)}  n_rs: {str(rs)}')
                        medscan.data.ROI.update_indexes(key=contour_num, indexes=np.NaN)
                    contour_num += 1

            # Save additional scan information in the MEDscan class
            medscan.data.set_patient_position(patient_position=dicom_h[0].PatientPosition)
            medscan.patientID = str(dicom_h[0].PatientID)
            medscan.format = "dicom"
            if 'SeriesDescription' in dicom_h[0]:
                medscan.series_description = dicom_h[0].SeriesDescription
            else:
                medscan.series_description = dicom_h[0].Modality

            # save MEDscan class instance as a pickle object
            if self.save and self.path_save:
                name_complete = save_MEDscan(medscan, self.path_save)
                del medscan
            else:
                series_description = medscan.series_description.translate({ord(ch): '-' for ch in '/\\ ()&:*'})
                name_id = medscan.patientID.translate({ord(ch): '-' for ch in '/\\ ()&:*'})

                # final saving name
                name_complete = name_id + '__' + series_description + '.' + medscan.type + '.npy'

        except Exception as e:
            if 'SeriesDescription' in dicom_hi[0]:
                print(f'patientID: {dicom_hi[0].PatientID} Modality: {dicom_hi[0].SeriesDescription} error: {str(e)}')
            else:
                print(f'patientID: {dicom_hi[0].PatientID} Modality: {dicom_hi[0].Modality} error: {str(e)}')
            return ''
        
        return name_complete



from . import *
from .batch_patients import *
from .create_radiomics_table import *
from .data_frame_export import *
from .find_process_names import *
from .get_file_paths import *
from .get_full_rad_names import *
from .get_institutions_from_ids import *
from .get_patient_id_from_scan_name import *
from .get_patient_names import *
from .get_radiomic_names import *
from .get_scan_name_from_rad_name import *
from .image_reader_SITK import *
from .image_volume_obj import *
from .imref import *
from .initialize_features_names import *
from .inpolygon import *
from .interp3 import *
from .json_utils import *
from .mode import *
from .parse_contour_string import *
from .save_MEDscan import *
from .strfind import *
from .textureTools import *
from .write_radiomics_csv import *



import os.path
from isort import file
import pandas as pd

def export_table(file_name: file,
                 data: object):
    """Export table

    Args:
        file_name (file): name of the file
        data (object): the data
    
    Returns:
        None
    """

    if not isinstance(data, (pd.DataFrame, pd.Series)):
        raise TypeError(f"The exported data should be a pandas DataFrame or Series. Found: {type(data)}")

    # Find the extension
    ext = os.path.splitext(file_name)[1]

    # Set an index switch based on type of input
    if isinstance(data, pd.DataFrame):
        write_index = False
    else:
        write_index = True

    if ext == ".csv":
        data.to_csv(path_or_buf=file_name, sep=";", index=write_index)
    elif ext in [".xls", ".xlsx"]:
        data.to_excel(excel_writer=file_name, index=write_index)
    elif ext in [".tex"]:
        with open(file=file_name, mode="w") as f:
            data.to_latex(buf=f, index=write_index)
    elif ext in [".html"]:
        with open(file=file_name, mode="w") as f:
            data.to_html(buf=f, index=write_index)
    elif ext in [".json"]:
        data.to_json(path_or_buf=file_name)
    else:
        raise ValueError(f"File extension not supported for export of table data. Recognised extensions are: \".csv\", \".xls\", \".xlsx\", \".tex\", \".html\" and \".json\". Found: {ext}")



import numpy as np
import pandas as pd
from neuroCombat import neuroCombat

from ..utils.get_institutions_from_ids import get_institutions_from_ids


class Normalization:
    def __init__(
            self, 
            method: str = 'combat', 
            variable_table: pd.DataFrame = None, 
            covariates_df: pd.DataFrame = None, 
            institutions: list = None
        ) -> None:
        """
        Constructor of the Normalization class.
        """
        self.method = method
        self.variable_table = variable_table
        self.covariates_df = covariates_df
        self.institutions = institutions
    
    def apply_combat(
            self, 
            variable_table: pd.DataFrame, 
            covariate_df: pd.DataFrame = None, 
            institutions: list = None
        ) -> pd.DataFrame:
        """
        Applys ComBat Normalization method to the data.
        More details :ref:`this link <https://github.com/Jfortin1/ComBatHarmonization/tree/master/Python>`.

        Args:
            variable_table (pd.DataFrame): pandas data frame on which Combat harmonization will be applied. 
                This table is of size N X F (Observations X Features) and has the IDs as index. 
                Requirements for this table

                    - Does not contain NaNs.
                    - No feature has 0 variance.
                    - All variables are continuous (For example: Radiomics variables).
            covariate_df (pd.DataFrame, optional): N X M pandas data frame, where N must equal the number of 
                observations in variable_table. M is the number of covariates to include in the algorithm.
            institutions (list, optional): List of size n_observations X 1 with the different institutions.
        
        Returns:
            pd.DataFrame: variable_table after Combat harmonization.
        """
        # Initializing the class attributes from the arguments
        if variable_table is None:
            if self.variable_table is None:
                raise ValueError('variable_table must be given.')
        else:
            self.variable_table = variable_table
        if covariate_df is not None:
            self.covariates_df = covariate_df
        if institutions:
            self.institutions = institutions
        
        # Intializing the institutions if not given
        if self.institutions is None:
            patient_ids = pd.Series(self.variable_table.index)
            self.institutions = get_institutions_from_ids(patient_ids)
            all_institutions = self.institutions.unique()
            for n in range(all_institutions.size):
                self.institutions[self.institutions == all_institutions[n]] = n+1
        self.institutions = self.institutions.to_numpy(dtype=int)
        self.institutions = np.reshape(self.institutions, (-1, 1))
        
        # No harmonization will be applied if there is only one institution
        if np.unique(self.institutions).size < 2:
            return self.variable_table
        
        # Initializing the covariates if not given
        if self.covariates_df is not None:
            self.covariates_df['institution'] = self.institutions
        else:
            # the covars matrix is only a row with the institution
            self.covariates_df = pd.DataFrame(
                self.institutions, 
                columns=['institution'], 
                index=self.variable_table.index.values
            )

        # Apply combat
        n_features = self.variable_table.shape[1]
        batch_col = 'institution'
        if n_features == 1:
            # combat does not work with a single feature so a temporary one is added,
            # then removed later (this has no effect on the algorithm).
            self.variable_table['temp'] = pd.Series(
                np.ones(self.variable_table.shape[0]),
                index=self.variable_table.index
            )
            data_combat = neuroCombat(
                self.variable_table.transpose(), 
                self.covariates_df, 
                batch_col
            )
            self.variable_table = pd.DataFrame(self.variable_table.drop('temp', axis=1))
            vt_combat = pd.DataFrame(data_combat[:][0].transpose())
        else:
            data_combat = neuroCombat(
                self.variable_table.transpose(), 
                self.covariates_df, 
                batch_col
            )
            vt_combat = pd.DataFrame(data_combat['data']).transpose()

        self.variable_table[:] = vt_combat.values

        return self.variable_table



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import logging
from copy import deepcopy
from typing import List

import numpy as np

from ..MEDscan import MEDscan
from ..processing.segmentation import compute_box
from ..utils.image_volume_obj import image_volume_obj
from ..utils.imref import imref3d, intrinsicToWorld, worldToIntrinsic
from ..utils.interp3 import interp3


def interp_volume(
        vol_obj_s: image_volume_obj,
        medscan: MEDscan= None, 
        vox_dim: List = None,
        interp_met: str = None,
        round_val: float = None,
        image_type: str = None,
        roi_obj_s: image_volume_obj = None,
        box_string: str = None,
        texture: bool = False) -> image_volume_obj:
    """3D voxel interpolation on the input volume.

    Args:
        vol_obj_s (image_volume_obj): Imaging  that will be interpolated.
        medscan (object): The MEDscan class object.
        vox_dim (array): Array of the voxel dimension. The following format is used
                         [Xin,Yin,Zslice], where Xin and Yin are the X (left to right) and
                         Y (bottom to top) IN-PLANE resolutions, and Zslice is the slice spacing,
                         no matter the orientation of the volume (i.e. axial , sagittal, coronal).
        interp_met (str): {nearest, linear, spline, cubic} optional, Interpolation method
        round_val (float): Rounding value. Must be between 0 and 1 for ROI interpolation
                           and to a power of 10 for Image interpolation.
        image_type (str): 'image' for imaging data interpolation and 'roi' for ROI mask data interpolation.
        roi_obj_s (image_volume_obj): Mask data, will be used to compute a new specific box
                                      and the new imref3d object for the imaging data.
        box_string (str): Specifies the size if the box containing the ROI

                          - 'full': full imaging data as output.
                          - 'box': computes the smallest bounding box.
                          - Ex: 'box10': 10 voxels in all three dimensions are added to \
                            the smallest bounding box. The number after 'box' defines the \
                            number of voxels to add.
                          - Ex: '2box': Computes the smallest box and outputs double its \
                            size. The number before 'box' defines the multiplication in size.
        texture (bool): If True, the texture voxel spacing of ``MEDscan`` will be used for interpolation.
    
    Returns:
        ndarray: 3D array of 1's and 0's defining the ROI mask.
    """
    try:
        # PARSING ARGUMENTS
        if vox_dim is None:
            if medscan is None:
                return deepcopy(vol_obj_s)
            else:
                if texture:
                    vox_dim = medscan.params.process.scale_text
                else:
                    vox_dim = medscan.params.process.scale_non_text
        if np.sum(vox_dim) == 0:
            return deepcopy(vol_obj_s)
        if len(vox_dim) == 2:
            two_d = True
        else:
            two_d = False
                
        if image_type is None:
            raise ValueError(
                "The type of input image should be specified as \"image\" or \"roi\".")
        elif image_type not in ["image", "roi"]:
            raise ValueError(
                "The type of input image should either be \"image\" or \"roi\".")
        elif image_type == "image":
            if not interp_met: 
                if medscan:
                    interp_met = medscan.params.process.vol_interp
                else:
                    raise ValueError("Interpolation method or MEDscan instance should be provided.")
            if interp_met not in ["linear", "cubic", "spline"]:
                raise ValueError(
                    "Interpolation method for images should either be \"linear\", \"cubic\" or \"spline\".")
            if medscan and not round_val:
                round_val = medscan.params.process.gl_round
            if round_val is not None:
                if np.mod(np.log10(round_val), 1):
                    raise ValueError("\"round_val\" should be a power of 10.")
        else:
            if not interp_met:
                if medscan:
                    interp_met = medscan.params.process.roi_interp
                else:
                    raise ValueError("Interpolation method or MEDscan instance should be provided.")
            if interp_met not in ["nearest", "linear", "cubic"]:
                raise ValueError(
                    "Interpolation method for images should either be \"nearest\", \"linear\" or \"cubic\".")
            if medscan and not round_val:
                round_val = medscan.params.process.roi_pv
            if round_val is not None:
                if round_val < 0.0 or round_val > 1.0:
                    raise ValueError("\"round_val\" must be between 0.0 and 1.0.")
            else:
                raise ValueError("\"round_val\" must be provided for \"roi\".")
        if medscan and not box_string:
            box_string = medscan.params.process.box_string
        if roi_obj_s is None or box_string is None:
            use_box = False
        else:
            use_box = True

        # --> QUERIED POINTS: NEW INTERPOLATED VOLUME: "q" or "Q".
        # --> SAMPLED POINTS: ORIGINAL VOLUME: "s" or "S".
        # --> Always using XYZ coordinates (unless specifically noted),
        #     not MATLAB IJK, so beware!

        # INITIALIZATION
        res_q = vox_dim
        if two_d:
            # If 2D, the resolution of the slice dimension of he queried volume is
            # set to the same as the sampled volume.
            res_q = np.concatenate((res_q, vol_obj_s.spatialRef.PixelExtentInWorldZ))

        res_s = np.array([vol_obj_s.spatialRef.PixelExtentInWorldX,
                        vol_obj_s.spatialRef.PixelExtentInWorldY,
                        vol_obj_s.spatialRef.PixelExtentInWorldZ])

        if np.array_equal(res_s, res_q):
            return deepcopy(vol_obj_s)

        spatial_ref_s = vol_obj_s.spatialRef
        extent_s = np.array([spatial_ref_s.ImageExtentInWorldX,
                            spatial_ref_s.ImageExtentInWorldY,
                            spatial_ref_s.ImageExtentInWorldZ])
        low_limits_s = np.array([spatial_ref_s.XWorldLimits[0],
                            spatial_ref_s.YWorldLimits[0],
                            spatial_ref_s.ZWorldLimits[0]])

        # CREATING QUERIED "imref3d" OBJECT CENTERED ON SAMPLED VOLUME

        # Switching to IJK (matlab) reference frame for "imref3d" computation.
        # Putting a "ceil", according to IBSI standards. This is safer than "round".
        size_q = np.ceil(np.around(np.divide(extent_s, res_q),
                                decimals=3)).astype(int).tolist()

        if two_d:
            # If 2D, forcing the size of the queried volume in the slice dimension
            # to be the same as the sample volume.
            size_q[2] = vol_obj_s.spatialRef.ImageSize[2]

        spatial_ref_q = imref3d(imageSize=size_q, 
                            pixelExtentInWorldX=res_q[0],
                            pixelExtentInWorldY=res_q[1],
                            pixelExtentInWorldZ=res_q[2])

        extent_q = np.array([spatial_ref_q.ImageExtentInWorldX,
                            spatial_ref_q.ImageExtentInWorldY,
                            spatial_ref_q.ImageExtentInWorldZ])
        low_limits_q = np.array([spatial_ref_q.XWorldLimits[0],
                            spatial_ref_q.YWorldLimits[0],
                            spatial_ref_q.ZWorldLimits[0]])
        diff = extent_q - extent_s
        new_low_limits_q = low_limits_s - diff/2
        spatial_ref_q.XWorldLimits = spatial_ref_q.XWorldLimits - \
            (low_limits_q[0] - new_low_limits_q[0])
        spatial_ref_q.YWorldLimits = spatial_ref_q.YWorldLimits - \
            (low_limits_q[1] - new_low_limits_q[1])
        spatial_ref_q.ZWorldLimits = spatial_ref_q.ZWorldLimits - \
            (low_limits_q[2] - new_low_limits_q[2])

        # REDUCE THE SIZE OF THE VOLUME PRIOR TO INTERPOLATION
        # TODO check that compute_box vol and roi are intended to be the same!
        if use_box:
            _, _, tempSpatialRef = compute_box(
                vol=roi_obj_s.data, roi=roi_obj_s.data, spatial_ref=vol_obj_s.spatialRef,
                box_string=box_string)

            size_temp = tempSpatialRef.ImageSize

            # Getting world boundaries (center of voxels) of the new box
            x_bound, y_bound, z_bound = intrinsicToWorld(R=tempSpatialRef,
                                                    xIntrinsic=np.array(
                                                        [0.0, size_temp[0]-1.0]),
                                                    yIntrinsic=np.array(
                                                        [0.0, size_temp[1]-1.0]),
                                                    zIntrinsic=np.array([0.0, size_temp[2]-1.0]))

            # Getting the image positions of the boundaries of the new box, IN THE
            # FULL QUERIED FRAME OF REFERENCE (centered on the sampled frame of
            # reference).
            x_bound, y_bound, z_bound = worldToIntrinsic(
                R=spatial_ref_q, xWorld=x_bound, yWorld=y_bound, zWorld=z_bound)

            # Rounding to the nearest image position integer
            x_bound = np.round(x_bound).astype(int)
            y_bound = np.round(y_bound).astype(int)
            z_bound = np.round(z_bound).astype(int)

            size_q = np.array([x_bound[1] - x_bound[0] + 1, y_bound[1] -
                            y_bound[0] + 1, z_bound[1] - z_bound[0] + 1])

            # Converting back to world positions ion order to correctly define
            # edges of the new box and thus center it onto the full queried
            # reference frame
            x_bound, y_bound, z_bound = intrinsicToWorld(R=spatial_ref_q,
                                                    xIntrinsic=x_bound,
                                                    yIntrinsic=y_bound,
                                                    zIntrinsic=z_bound)

            new_low_limits_q[0] = x_bound[0] - res_q[0]/2
            new_low_limits_q[1] = y_bound[0] - res_q[1]/2
            new_low_limits_q[2] = z_bound[0] - res_q[2]/2

            spatial_ref_q = imref3d(imageSize=size_q, 
                                pixelExtentInWorldX=res_q[0],
                                pixelExtentInWorldY=res_q[1],
                                pixelExtentInWorldZ=res_q[2])

            spatial_ref_q.XWorldLimits -= spatial_ref_q.XWorldLimits[0] - \
                new_low_limits_q[0]
            spatial_ref_q.YWorldLimits -= spatial_ref_q.YWorldLimits[0] - \
                new_low_limits_q[1]
            spatial_ref_q.ZWorldLimits -= spatial_ref_q.ZWorldLimits[0] - \
                new_low_limits_q[2]

        # CREATING QUERIED XYZ POINTS
        x_q = np.arange(size_q[0])
        y_q = np.arange(size_q[1])
        z_q = np.arange(size_q[2])
        x_q, y_q, z_q = np.meshgrid(x_q, y_q, z_q, indexing='ij')
        x_q, y_q, z_q = intrinsicToWorld(
            R=spatial_ref_q, xIntrinsic=x_q, yIntrinsic=y_q, zIntrinsic=z_q)

        # CONVERTING QUERIED XZY POINTS TO INTRINSIC COORDINATES IN THE SAMPLED
        # REFERENCE FRAME
        x_q, y_q, z_q = worldToIntrinsic(
            R=spatial_ref_s, xWorld=x_q, yWorld=y_q, zWorld=z_q)

        # INTERPOLATING VOLUME
        data = interp3(v=vol_obj_s.data, x_q=x_q, y_q=y_q, z_q=z_q, method=interp_met)
        vol_obj_q = image_volume_obj(data=data, spatial_ref=spatial_ref_q)

        # ROUNDING
        if image_type == "image":
            # Grey level rounding for "image" type
            if round_val is not None and (type(round_val) is int or type(round_val) is float):
                # DELETE NEXT LINE WHEN THE RADIOMICS PARAMETER OPTIONS OF
                # interp.glRound ARE FIXED
                round_val = (-np.log10(round_val)).astype(int)
                vol_obj_q.data = np.around(vol_obj_q.data, decimals=round_val)
        else:
            vol_obj_q.data[vol_obj_q.data >= round_val] = 1.0
            vol_obj_q.data[vol_obj_q.data < round_val] = 0.0

    except Exception as e:
        if medscan:
            if medscan.params.radiomics.scale_name:
                message = f"\n PROBLEM WITH INTERPOLATION:\n {e}"
                logging.error(message)
                medscan.radiomics.image.update(
                    {(medscan.params.radiomics.scale_name ): 'ERROR_PROCESSING'})
            else:
                message = f"\n PROBLEM WITH INTERPOLATION:\n {e}"
                logging.error(message)
                medscan.radiomics.image.update(
                    {('scale'+(str(medscan.params.process.scale_non_text[0])).replace('.','dot')): 'ERROR_PROCESSING'})
        else:
            print(f"\n PROBLEM WITH INTERPOLATION:\n {e}")

    return vol_obj_q
